
from typing import Dict

import traceback
import math
import datetime
import time
import numpy as np
import pandas as pd

from ...manager.manager_fund import FundDataManager
from ...view.derived_models import FundIndicatorMonthly


class FundIndicatorProcessorMonthly:

    _RF = 0.011

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, start_date, end_date, dm=None):
        if dm is None:
            self._dm = FundDataManager(start_time=start_date, end_time=end_date)
            self._dm.init(score_pre_calc=False)
        else:
            self._dm = dm
        self.start_date = self._dm.start_date
        self.end_date = self._dm.end_date

        # index price, fund nav 日期对齐至交易日
        self._trading_days = self._dm.dts.trading_days[(self._dm.dts.trading_days.datetime >= self.start_date) & (self._dm.dts.trading_days.datetime <= self.end_date)].datetime
        self._fund_nav: pd.DataFrame = self._dm.dts.fund_nav.reindex(self._trading_days)
        self._index_price: pd.DataFrame = self._dm.dts.index_price.reindex(self._trading_days)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)

        # 过滤出来需要用到的fund_info
        self._fund_info: pd.DataFrame = self._dm.dts.fund_info[self._dm.dts.fund_info['fund_id'].isin(self._fund_nav.columns)]

        # 获取fund到index的字典
        fund_to_index: Dict[str, str] = self._fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']

        # 超过基金终止日的基金净值赋空
        fund_to_end_date_dict = self._fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        for fund_id in self._fund_nav.columns:
            fund_end_date = fund_to_end_date_dict[fund_id]
            if self.end_date >= fund_end_date:
                self._fund_nav.loc[fund_end_date:, fund_id] = np.nan

        self._index_ret: pd.DataFrame = self._get_return(self._index_price)
        self._fund_ret: pd.DataFrame = self._get_return(self._fund_nav)
        index_list: pd.Series = self._fund_ret.columns.map(fund_to_index)
        index_list = index_list[index_list.notna()]
        self._index_list: np.ndarray = index_list.unique()

        # resample
        self._index_ret_resampled: pd.DataFrame = self._get_resample_ret(self._index_ret)
        # self._national_debt_ret_resampled: pd.DataFrame = self._index_ret_resampled.loc[:, 'national_debt']
        self._fund_ret_resampled: pd.DataFrame = self._get_resample_ret(self._fund_ret)

        self._index_fund: pd.DataFrame = self._fund_info.loc[self._fund_info.index_id.isin(self._index_list), ['index_id', 'fund_id']].groupby('index_id').apply(lambda x: x['fund_id'].to_list())
        self._fund_start_date: Dict[str] = self._fund_info[['fund_id', 'start_date']].set_index('fund_id').to_dict()['start_date']

    def _get_return(self, df) -> pd.DataFrame:
        ret = df.applymap(np.log).diff()
        return ret.iloc[1:, :]

    def _get_resample_ret(self, df, rule='1M'):
        return df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum()

    def _process_fund_indicator_update(self, index_id, index_ret: pd.Series, index_ret_resampled: pd.Series,
                                       fund_list, fund_ret: pd.DataFrame, fund_ret_resampled: pd.DataFrame) -> pd.DataFrame:
        # beta
        beta = fund_ret_resampled.apply(lambda x: index_ret_resampled.cov(x)) / index_ret_resampled.var()

        # sharpe ratio
        annualized_ret_M = fund_ret_resampled.sum() * (12 / len(fund_ret_resampled)) - self._RF
        annualized_vol_M = fund_ret_resampled.std(ddof=0) * math.sqrt(12)
        sharpe_ratio_M = annualized_ret_M / annualized_vol_M

        # treynor ratio
        # ex_ret = fund_ret_resampled.sub(self._national_debt_ret_resampled, axis=0)
        ex_ret = fund_ret_resampled.sub(self._RF / 12, axis=0)
        ex_ret_sum = ex_ret.sum()
        annualized_ret_ex = ex_ret_sum * 12 / len(fund_ret_resampled)
        treynor_ratio = annualized_ret_ex / beta

        # information ratio
        annualized_vol_ex = fund_ret_resampled.std() * math.sqrt(12)
        information_ratio = annualized_ret_ex / annualized_vol_ex

        # jensen alpha
        ri = (fund_ret_resampled.sum()) * 12 / len(fund_ret_resampled)
        rm = (index_ret_resampled.sum()) * 12 / len(index_ret_resampled)

        # rf = (self._national_debt_ret_resampled.sum()) * 12 / len(self._national_debt_ret_resampled)
        jensen_alpha = ri - self._RF - beta * (rm - self._RF)

        # mdd
        net_val = self._fund_nav.loc[:, fund_list]
        peak_series = net_val.rolling(window=len(net_val), min_periods=1).max()
        max_drawdown = (net_val / peak_series - 1.0).min()

        # calmar ratio
        calmar_ratio = annualized_ret_ex / max_drawdown

        return pd.DataFrame({'beta_m': beta, 'sharpe_ratio_m': sharpe_ratio_M,
                             'treynor_ratio_m': treynor_ratio, 'information_ratio_m': information_ratio,
                             'jensen_alpha_m': jensen_alpha, 'calmar_ratio_m': calmar_ratio, })

    def _calculate_update(self):
        result = []
        for index_id in self._index_list:
            _tm_1 = time.time()
            fund_list = self._index_fund[index_id]
            fund_list = list(set(fund_list) & set(self._fund_ret.columns))
            fund_ret = self._fund_ret.loc[:, fund_list]
            fund_ret_resampled = self._fund_ret_resampled.loc[:, fund_list]
            index_ret = self._index_ret.loc[:, index_id]
            index_ret_resampled = self._index_ret_resampled.loc[:, index_id]
            result.append(self._process_fund_indicator_update(index_id, index_ret, index_ret_resampled, fund_list, fund_ret, fund_ret_resampled))
            _tm_2 = time.time()
            print(f'index {index_id} fund number : {len(fund_list)} finish, cost time {_tm_2 - _tm_1}s')
        self.result = pd.concat(result)
        self.result = self.result.replace({-np.Inf:None,np.Inf:None})

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date):
        # 月更因子, 由于resample出来的日期都是每月最后一天，所以每月最后一天前的最后一个交易日更新
        failed_tasks = []
        try:
            print(f'monthly indicator update on the last day of month {end_date_dt}')
            start_date_dt = pd.to_datetime(start_date, infer_datetime_format=True).date()
            start_date = start_date_dt - datetime.timedelta(days=1150)  # 3年历史保险起见，多取几天 3*365=1095
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
            self.init(start_date=start_date, end_date=end_date)
            self._calculate_update()
            self._data_helper._upload_derived(self.result.assign(datetime=end_date_dt).reset_index(), FundIndicatorMonthly.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_indicator_monthly')
        return failed_tasks



import traceback
import json
from ...manager.manager_fund import *
from ...view.derived_models import *
from .derived_data_helper import DerivedDataHelper

class TopManagerFundProcessor(object):

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, start_date='20100101', end_date='20200930'):
        self.start_date = start_date
        self.end_date = end_date
        self.white_list = pd.read_csv('../rpm/etc/chuang_jin_modify.csv').fund_id.tolist()

        with BasicDatabaseConnector().managed_session() as quant_session:
            query = quant_session.query(
                TradingDayList
            )
            trading_days = pd.read_sql(query.statement, query.session.bind)
            self.trading_dates = trading_days[trading_days['datetime'] >= self.start_date].datetime

        with DerivedDatabaseConnector().managed_session() as session:        
            query = session.query(
                FundManagerScore
            ).filter(
                FundManagerScore.fund_type == 'stock',
                FundManagerScore.datetime >= self.start_date,
                FundManagerScore.datetime <= self.end_date,
            )
            self.mng_score_df = pd.read_sql(query.statement, query.session.bind)
            self.mng_score_df = self.mng_score_df.set_index(['datetime','manager_id']).drop(columns='fund_type')

            query = session.query(
                FundManagerFundRank
            ).filter(
                FundManagerFundRank.datetime >= self.start_date,
                FundManagerFundRank.datetime <= self.end_date,
            )
            self.mng_fund_list = pd.read_sql(query.statement, query.session.bind)
            self.mng_fund_list = self.mng_fund_list.set_index(['datetime','mng_id'])
        
    def calculate_update(self):
        res = []
        for dt in self.trading_dates:
            try:
                top_mng_list = self.mng_score_df.loc[dt].sort_values('score',ascending=False).index.tolist()
                mng_fund_dt = self.mng_fund_list.loc[dt]
                mng_fund_dt_list = mng_fund_dt.index.tolist()
                top_mng_list = [_ for _ in top_mng_list if (_ in mng_fund_dt_list)]
                top_funds = [json.loads(_)[0] for _ in mng_fund_dt.loc[top_mng_list].fund_list.values]
                top_funds = [ _ for _ in top_funds if _ in self.white_list][:10]
                dic = {
                    'datetime':dt,
                }
                for idx, i in enumerate(top_funds):
                    dic.update({f'top_{idx+1}':i})
                res.append(dic)
            except:
                continue
        self.result = pd.DataFrame(res)

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            self.init(start_date=start_date_dt, end_date=end_date_dt)
            self.calculate_update()
            df = self.result[self.result['datetime'] == end_date_dt]
            self._data_helper._upload_derived(df, TopManagerFunds.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('top_manager_funds')

        return failed_tasks
        



import time
from typing import Tuple, Set, Optional, List
from collections import defaultdict

import pandas as pd

from ...api.raw import RawDataApi
from ...wrapper.mysql import RawDatabaseConnector

ALL_CSI_INDEX_INNER_NAME = '_csi_index_all'


class RawDataHelper():
    def __init__(self):
        self._updated_count = defaultdict(int)
        self._timer: float = time.monotonic()

    def _upload_raw(self, df, table_name, to_truncate=False):
        print(table_name)
        # print(df)
        if to_truncate:
            with RawDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')

        df = df.drop(columns='_update_time', errors='ignore')
        df.to_sql(table_name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

        now: float = time.monotonic()
        print(f'{table_name} costs {now - self._timer}s')
        self._timer = now
        self._updated_count[table_name] += df.shape[0]

    @staticmethod
    def get_new_and_delisted_fund_list(date: str) -> Tuple[Set[str], Set[str]]:
        fund_list = RawDataApi().get_em_fund_list(date, limit=2)
        if fund_list is None or fund_list.shape[0] < 2:
            print(f"failed to get two days' fund list, (date){date}")
            return (None, None)

        lastest_list: pd.Series = fund_list.iloc[0, :]
        last_list: pd.Series = fund_list.iloc[1, :]
        print(f"got two days' fund list, (latest){lastest_list.datetime} (prev){last_list.datetime}")
        return (set(lastest_list.all_live_fund_list.split(',')).difference(set(last_list.all_live_fund_list.split(','))),
                set(lastest_list.delisted_fund_list.split(',')).difference(set(last_list.delisted_fund_list.split(','))))

    @staticmethod
    def get_indexes_that_become_invalid(end_date: str) -> Set[str]:
        index_component = RawDataApi().get_em_index_component(end_date=end_date, index_list=[ALL_CSI_INDEX_INNER_NAME])
        if index_component is None or index_component.shape[0] < 2:
            print(f'failed to get em index component, (end_date){end_date}')
            return

        lastest_list: pd.Series = index_component.iloc[-1, :]
        last_list: pd.Series = index_component.iloc[-2, :]
        print(f"got two days' index component, (latest){lastest_list.datetime} (prev){last_list.datetime}")
        return set(lastest_list.stock_list.split(',')).difference(set(last_list.stock_list.split(',')))

    @staticmethod
    def get_all_live_fund_list(end_date: str) -> Optional[List[str]]:
        fund_list_df = RawDataApi().get_em_fund_list(end_date, limit=1)
        if fund_list_df is None:
            return
        return fund_list_df.all_live_fund_list.array[0].split(',')

    @staticmethod
    def get_all_fund_list(end_date: str) -> Optional[List[str]]:
        fund_list_df = RawDataApi().get_em_fund_list(end_date, limit=1)
        if fund_list_df is None:
            return
        fund_list = fund_list_df.all_live_fund_list.array[0].split(',')
        fund_list += fund_list_df.delisted_fund_list.array[0].split(',')
        return fund_list



import pandas as pd
import traceback
import datetime
from collections import Iterable
from sqlalchemy import distinct, func
from cassandra.cqlengine.query import BatchQuery
from ...wrapper.mysql import DerivedDatabaseConnector, BasicDatabaseConnector, ViewDatabaseConnector
from ...view.view_models import IndexDailyCollection
from ...view.basic_models import IndexInfo, IndexPrice, IndexComponent
from ...view.derived_models import IndexValuationLongTerm, IndexReturn, IndexVolatility
from ....util.calculator import Calculator as SurfingCalculator
from ...view.cas.tag_method_index_level import TagMethodIndexLevel
from ....data.api.derived import DerivedDataApi
from ....constant import SectorType

class IndexDailyCollectionProcessor(object):
    def get_index_volatility(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexVolatility).order_by(IndexVolatility.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexVolatility.index_id,
                    IndexVolatility.datetime,
                    IndexVolatility.w1_vol,
                    IndexVolatility.m1_vol,
                    IndexVolatility.m3_vol,
                    IndexVolatility.m6_vol,
                    IndexVolatility.y1_vol,
                    IndexVolatility.y3_vol,
                    IndexVolatility.y5_vol,
                    IndexVolatility.y10_vol,
                    IndexVolatility.this_y_vol,
                    IndexVolatility.cumulative_vol,
                ).filter(IndexVolatility.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'vol_datetime'})
                df = df.set_index('index_id')
                return df
            except Exception as e:
                print('Failed get_index_volatility <err_msg> {}'.format(e))

    # TODO: Use basic.index_valuation_develop table
    def get_index_valuation(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexValuationLongTerm).order_by(IndexValuationLongTerm.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexValuationLongTerm.index_id,
                    IndexValuationLongTerm.pb_mrq,
                    IndexValuationLongTerm.pe_ttm,
                    IndexValuationLongTerm.peg_ttm,
                    IndexValuationLongTerm.roe,
                    IndexValuationLongTerm.dy,
                    IndexValuationLongTerm.pe_pct,
                    IndexValuationLongTerm.pb_pct,
                    IndexValuationLongTerm.ps_pct,
                    IndexValuationLongTerm.val_score,
                    IndexValuationLongTerm.datetime,
                    IndexValuationLongTerm.est_peg,
                ).filter(IndexValuationLongTerm.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_valuation <err_msg> {}'.format(e))

    def get_index_return(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexReturn).order_by(IndexReturn.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexReturn.index_id,
                    IndexReturn.datetime,
                    IndexReturn.w1_ret,
                    IndexReturn.m1_ret,
                    IndexReturn.m3_ret,
                    IndexReturn.m6_ret,
                    IndexReturn.y1_ret,
                    IndexReturn.y3_ret,
                    IndexReturn.y5_ret,
                    IndexReturn.y10_ret,
                    IndexReturn.this_y_ret,
                    IndexReturn.cumulative_ret,
                ).filter(IndexReturn.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'ret_datetime'})
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_return<err_msg> {}'.format(e))

    def get_index_info(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexInfo.index_id,
                    IndexInfo.em_id,
                    IndexInfo.industry_tag,
                    IndexInfo.tag_method,
                    IndexInfo.desc_name,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_info<err_msg> {}'.format(e))

    def get_index_component(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexComponent.index_id,
                    IndexComponent.id_cat,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df['id_cat'] = df['id_cat'].transform(lambda x: SectorType(x).name)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_info<err_msg> {}'.format(e))

    def get_index_price(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                subq = mn_session.query(
                    IndexPrice.index_id.label('temp_id'),
                    func.max(IndexPrice.datetime).label('temp_date'),
                ).group_by(IndexPrice.index_id).subquery()
                query = mn_session.query(
                    IndexPrice.index_id,
                    IndexPrice.datetime,
                    IndexPrice.volume,
                    IndexPrice.low,
                    IndexPrice.close,
                    IndexPrice.high,
                    IndexPrice.open,
                    IndexPrice.ret,
                    IndexPrice.total_turnover
                ).filter(
                    IndexPrice.index_id == subq.c.temp_id,
                    IndexPrice.datetime == subq.c.temp_date
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'price_datetime'})
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_price<err_msg> {}'.format(e))

    def calc_percentile(self, df):
        def helper(x, temp_df, column):
            temp_df = temp_df[temp_df['index_id'] == x]
            ret = temp_df[column].quantile([0, .3, .7, 1])
            return ret

        start_date = datetime.datetime.now().date() - datetime.timedelta(days=3700)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuationLongTerm.pe_ttm,
                IndexValuationLongTerm.pb_mrq,
                IndexValuationLongTerm.index_id
            ).filter(
                IndexValuationLongTerm.datetime >= start_date,
            )
            pe_pb_df = pd.read_sql(query.statement, query.session.bind)

        dff = df['index_id'].apply(helper, args=(pe_pb_df, 'pe_ttm',))
        df['pe_0_percentile'] = dff[0.0]
        df['pe_30_percentile'] = dff[0.3]
        df['pe_70_percentile'] = dff[0.7]
        df['pe_100_percentile'] = dff[1.0]

        dff = df['index_id'].apply(helper, args=(pe_pb_df, 'pb_mrq',))
        df['pb_0_percentile'] = dff[0.0]
        df['pb_30_percentile'] = dff[0.3]
        df['pb_70_percentile'] = dff[0.7]
        df['pb_100_percentile'] = dff[1.0]
        return df

    def get_index_yesterday_price(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                yest_time = mn_session.query(
                    distinct(IndexPrice.datetime)
                ).order_by(
                    IndexPrice.datetime.desc()
                ).offset(1).limit(1).one_or_none()
                yest_time = yest_time[0]
                print(yest_time)
                query = mn_session.query(
                    IndexPrice.index_id,
                    IndexPrice.close,
                ).filter(IndexPrice.datetime==yest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={
                    'close': 'yest_close',
                })
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_price<err_msg> {}'.format(e))

    def index_level_save_to_cas(self, items):
        groups = [i for i in range(0, len(items), 100)]
        for i in groups:
            if i != groups[-1]:
                end = i + 100
            else:
                end = len(items)
            for item in items[i: end]:
                with BatchQuery() as batch:
                    TagMethodIndexLevel.batch(batch).create(
                        index_id=item.get('index_id'),
                        data=item.get('data'),
                        update_time=datetime.datetime.now(),
                    )

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def replace_datetime(self, obj):
        if type(obj) == str:
            return obj

        if type(obj) == dict:
            r = {}
            for key in obj:
                r[key] = self.replace_datetime(obj[key])
            return r

        if isinstance(obj, Iterable):
            return list(map(lambda x: self.replace_datetime(x), obj))

        if isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')

        return obj

    def collection_daily_index(self):
        try:
            print('1、 load data...')
            info = self.get_index_info()
            vol = self.get_index_volatility().reindex(info.index)
            ret = self.get_index_return().reindex(info.index)
            val = self.get_index_valuation().reindex(info.index)
            price = self.get_index_price().reindex(info.index)
            yest_price = self.get_index_yesterday_price().reindex(info.index)
            component = self.get_index_component().reindex(info.index)

            print('2、 concat data...')
            df = pd.concat([info, vol, ret, val, price, yest_price, component], axis=1, sort=False)
            df.index.name = 'index_id'
            df = df.reset_index()

            print('3、 calculate tag_method index level')
            items = []
            for i in df.index:
                if not df.loc[i, 'tag_method']:
                    d = []
                else:
                    d = SurfingCalculator().get_index_level_index_by_tag_method(
                        df.loc[i, 'index_id'],
                        df.loc[i, 'tag_method'],
                    )
                items.append({
                    'index_id': df.loc[i, 'index_id'],
                    'data': self.replace_datetime(d)
                })

            self.index_level_save_to_cas(items)

            print('4、 special option...')

            df['order_book_id'] = df['em_id'].apply(lambda x: x.split('.')[0] if (x and x != 'not_available') else None)
            df = df.drop(['em_id'], axis=1)
            df['industry_tag'] = df['industry_tag'].apply(lambda x: x if x != 'not_available' else None)
            df = self.calc_percentile(df)

            self.append_data(IndexDailyCollection.__tablename__, df)

            print(df)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.collection_daily_index():
            failed_tasks.append('collection_daily_index')
        return failed_tasks


if __name__ == '__main__':
    IndexDailyCollectionProcessor().collection_daily_index()



import pandas as pd

from ...util.singleton import Singleton
from ..wrapper.mysql import ViewDatabaseConnector
from ..view.view_models import *


class ViewDataApi(metaclass=Singleton):
    @staticmethod
    def get_fund_daily_collection():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(FundDailyCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=FundDailyCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_daily_index_collection():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(IndexDailyCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=IndexDailyCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False

    @staticmethod
    def get_daily_fund_automatic_investment():
        with ViewDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(AutomaticInvestmentCollection)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns=AutomaticInvestmentCollection.trans_columns())
                return df
            except Exception as e:
                print(e)
                return False




import pandas as pd
import numpy as np
from datetime import datetime
from datetime import date
import json
from ...wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi

class IndexValuationProcessor():
    w1 = 5
    m1 = 20
    m3 = 60
    m6 = 121
    y1 = 242
    y3 = 242*2
    y5 = 242*5
    y10 = 242*10

    def __init__(self):
        self._raw_data_api = RawDataApi()
        self._basic_data_api = BasicDataApi()

    def upload_derived(self, df, table_name):
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
    
    def alpha(self, fund_nav,d):
        return fund_nav['adjusted_net_value'] / fund_nav['adjusted_net_value'].shift(d)  - fund_nav['close'] / fund_nav['close'].shift(d)

    def index_return(self):
        #history
        index_df = self._basic_data_api.get_index_info()
        index_list = [i for o,i in zip(index_df['order_book_id'],index_df['index_id']) if '.' in o]
        for i in index_list:
            
            df =  self._basic_data_api.get_index_price([i])
            if df.empty:
                continue
            else:
                df['w1_ret'] = df['close'] / df['close'].shift(self.w1) -1
                df['m1_ret'] = df['close'] / df['close'].shift(self.m1) -1
                df['m3_ret'] = df['close'] / df['close'].shift(self.m3) -1
                df['m6_ret'] = df['close'] / df['close'].shift(self.m6) -1
                df['y1_ret'] = df['close'] / df['close'].shift(self.y1) -1
                df['y3_ret'] = df['close'] / df['close'].shift(self.y3) -1
                df['y5_ret'] = df['close'] / df['close'].shift(self.y5) -1
                df['y10_ret'] = df['close'] / df['close'].shift(self.y10) -1
                df['cumulative_ret'] = df['close'] / df['close'].values[0] -1
                res = []
                df = df.set_index('datetime')
                for i,d in zip(df.close, df.index):
                    begin_of_year = df.index[df.index >= date(d.year, 1, 1)].min()
                    res.append(i/df.loc[begin_of_year,'close'] - 1)
                df['this_y_ret'] = res
                df = df.drop(['id','volume','low','close','high','open','total_turnover','ret'], axis = 1)
                df['datetime'] = df.index
                self.upload_derived(df, 'index_return')
    
    def index_volatility(self):
        #history
        index_df = self._basic_data_api.get_index_info()
        index_list = index_df['index_id'].tolist()
        for i in index_list:
            df =  self._basic_data_api.get_index_price([i])
            if df.empty:
                continue
            else:
                df = df.set_index('datetime')
                df = df[['close']]
                df = pd.DataFrame(df['close'] / df['close'][0])
                df['w1_vol'] = df['close'].rolling(self.w1).std()
                df['m1_vol'] = df['close'].rolling(self.m1).std()
                df['m3_vol'] = df['close'].rolling(self.m3).std()
                df['m6_vol'] = df['close'].rolling(self.m6).std()
                df['y1_vol'] = df['close'].rolling(self.y1).std()
                df['y3_vol'] = df['close'].rolling(self.y3).std()
                df['y5_vol'] = df['close'].rolling(self.y5).std()
                df['y10_vol'] = df['close'].rolling(self.y10).std()

                cumu_vol = []
                this_y_vol = []
                for d in df.index:
                    begin_of_year = df.index[df.index >= date(d.year, 1, 1)].min()
                    this_y_vol.append(df.loc[begin_of_year:d,]['close'].std())
                    cumu_vol.append(df.loc[:d,]['close'].std())
                df['this_y_vol'] = this_y_vol
                df['cumulative_vol'] = cumu_vol
                df = df.drop(['close'], axis = 1)
                df['index_id'] = i
                df['datetime'] = df.index
                self.upload_derived(df, 'index_volatility')
    
    def fund_alpha(self):

        index_df = self._basic_data_api.get_index_info()
        fund_info = self._basic_data_api.get_fund_info()
        fund_dic = {f: i for f,i in zip(fund_info['fund_id'],fund_info['track_index']) if i != 'none'}
        for f,i in fund_dic.items():

            fund_nav = self._basic_data_api.get_fund_nav([f]).drop_duplicates(subset='datetime')
            index = self._basic_data_api.get_index_price([i])

            fund_nav = fund_nav[['adjusted_net_value','datetime']].set_index('datetime')
            index = index[['close','datetime']].set_index('datetime')
            fund_nav = fund_nav.join(index)

            fund_nav['fund_ret'] = fund_nav['adjusted_net_value']/ fund_nav['adjusted_net_value'].shift(1) - 1
            fund_nav['index_ret'] = fund_nav['close'] /  fund_nav['close'].shift(1) - 1 
            track_err = []
            this_y_alpha = []
            cumu_alpha = []
            for d in fund_nav.index:
                f_ret = fund_nav.loc[:d,'fund_ret']
                i_ret = fund_nav.loc[:d,'index_ret']
                t = fund_nav.index.tolist().index(d) +1
                track_err_i = np.sqrt(pow((f_ret - i_ret),2).sum() *self.y1/ t)
                track_err.append(track_err_i)

                begin_of_year = fund_nav.index[fund_nav.index >= date(d.year, 1, 1)].min()
                this_y_alpha_i = fund_nav.loc[d,'adjusted_net_value'] / fund_nav.loc[begin_of_year,'adjusted_net_value'] - fund_nav.loc[d,'close'] / fund_nav.loc[begin_of_year,'close']
                this_y_alpha.append(this_y_alpha_i)

                cumu_alpha_i = fund_nav.loc[d,'adjusted_net_value'] / fund_nav['adjusted_net_value'].values[0] - fund_nav.loc[d,'close'] / fund_nav['close'].values[0]
                cumu_alpha.append(cumu_alpha_i)

            fund_nav['track_err'] = track_err
            fund_nav['this_y_alpha'] = this_y_alpha
            fund_nav['cumulative_alpha'] = cumu_alpha

            fund_nav['w1_alpha'] = self.alpha(fund_nav,self.w1)
            fund_nav['m1_alpha'] = self.alpha(fund_nav,self.m1)
            fund_nav['m3_alpha'] = self.alpha(fund_nav,self.m3)
            fund_nav['m6_alpha'] = self.alpha(fund_nav,self.m6)
            fund_nav['y1_alpha'] = self.alpha(fund_nav,self.y1)
            fund_nav['y3_alpha'] = self.alpha(fund_nav,self.y3)
            fund_nav['y5_alpha'] = self.alpha(fund_nav,self.y5)
            fund_nav['y10_alpha'] = self.alpha(fund_nav,self.y10)
            fund_nav = fund_nav.drop(['adjusted_net_value','close','fund_ret','index_ret'], axis = 1)
            fund_nav['fund_id'] = f
            fund_nav['datetime'] = fund_nav.index
            self.upload_derived(fund_nav, 'fund_alpha')
            

    def index_valuation(self):
        index_df = self._basic_data_api.get_index_info()
        index_df = index_df[index_df['tag_method'] != 'none']
        index_df = index_df[index_df['order_book_id'] != 'not_available']
        index_df = index_df[index_df['order_book_id'].map(lambda x : '.' in x )]
        start_date = date(2010,1,1)
        index_result = []
        for i in index_df.to_dict('records'):
            index_res = {}
            index_id = i['index_id']
            index_weight = self._raw_data_api.get_rq_index_weight([index_id],start_date)
            print(index_id)
            # save index components weight
            res = {}
            stock_list = []
            for s,w,d  in zip(index_weight['stock_list'] , index_weight['weight_list'],index_weight['datetime']):
                s = json.loads(s)
                w = json.loads(w)
                res[d] = [{'stock_id':si , 'weights': wi} for si,wi in zip(s,w)]
                stock_list.extend(s)
            stock_list = list(set(stock_list))
            # if all components are not stock, next one 
            stock_list = [ _ for _ in stock_list if _[0] in ['0','3','6']]
            if len(stock_list) < 1:
                continue
                
            stock_val  = self._raw_data_api.get_rq_stock_valuation(stock_list,start_date)
            stock_val2 = self._raw_data_api.get_stock_fin_fac(stock_list,start_date)
            
            for d, ws in res.items():
                index_res[d] = {}
                stock_val_tmp = stock_val[stock_val['datetime'] == d].copy()
                stock_val_tmp2 = stock_val2[stock_val2['datetime'] == d].copy()
                
                if stock_val_tmp.empty:
                    break
                for index_i in ['pb_ratio_lf','pe_ratio_ttm','peg_ratio_ttm','dividend_yield_ttm']:
                
                    index_d_w = stock_val_tmp[['stock_id',index_i]].set_index('stock_id').join(pd.DataFrame(res[d]).set_index('stock_id'))
                    index_d_w = index_d_w.fillna(0)
                    index_res[d][index_i] =  (index_d_w[index_i] * index_d_w['weights']).sum()
                
                index_d_w = stock_val2[['stock_id','return_on_equity_ttm']].set_index('stock_id').join(pd.DataFrame(res[d]).set_index('stock_id'))
                index_d_w = index_d_w.fillna(0)
                index_res[d]['return_on_equity_ttm'] = (index_d_w['return_on_equity_ttm'] * index_d_w['weights']).sum()
                
                
            index_result_i = pd.DataFrame(index_res).T
            index_result_i['index_id'] = index_id
            index_result_i['pe_pct'] = index_result_i['pe_ratio_ttm'].rank(pct=True)
            index_result_i['pb_pct'] = index_result_i['pb_ratio_lf'].rank(pct=True)
            index_result_i = index_result_i.dropna().copy()
            index_i = index_result_i['index_id'].values[0]
            val_method = index_df[index_df['index_id'] == index_i]['tag_method'].values[0]
            if val_method == 'PE百分位':
                index_result_i['val_score'] = index_result_i['pe_pct'].copy()
            else:
                index_result_i['val_score'] = index_result_i['pb_pct'].copy()
            index_result_i['datetime'] = index_result_i.index
            index_result_i.columns = ['dy_ttm','pb_mrq','pe_ttm','peg_ttm','roe_ttm','index_id','pe_pct','pb_pct','val_score','datetime']
            self.upload_derived(index_result_i, 'index_valuation')
            print(index_id, 'finish')

if __name__ == '__main__':
    ivp = IndexValuationProcessor()
    ivp.index_valuation()



import time
from typing import Union, Tuple

import numpy as np
import statsmodels.api as sm
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from collections import defaultdict
from ...wrapper.mysql import DerivedDatabaseConnector


FUND_CLASSIFIER = {
    'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金'],
    'stock': ['偏股混合型基金', '普通股票型基金', '平衡混合型基金', '灵活配置型基金'],  # 股票多空 对应绝对收益基金
    'bond': ['混合债券型二级基金', '中长期纯债型基金', '混合债券型一级基金', '短期纯债型基金', '偏债混合型基金', '平衡混合型基金', '灵活配置型基金'],
    'mmf': ['货币市场型基金'],
    'mix': ['平衡混合型基金', '灵活配置型基金']
}


def normalize(a: Union[pd.DataFrame, pd.Series]) -> np.ndarray:
    if isinstance(a, pd.Series):
        a = pd.DataFrame(a)
    return StandardScaler().fit_transform(a)


def score_rescale(a: Union[pd.DataFrame, pd.Series, np.ndarray]) -> np.ndarray:
    if isinstance(a, pd.Series):
        a = pd.DataFrame(a)
    return MinMaxScaler().fit_transform(a) * 100


def cont_rescale(a: Union[pd.DataFrame, pd.Series], feature_range: Tuple[float, float] = (0, 1)) -> np.ndarray:
    if isinstance(a, pd.Series):
        a = pd.DataFrame(a)
    return MinMaxScaler(feature_range=feature_range).fit_transform(a)


class DerivedDataHelper:
    def __init__(self):
        self._updated_count = defaultdict(int)
        self._timer = time.monotonic()

    @staticmethod
    def _lambda_cl(total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 0], total[:, 0]]).T
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X[:, 0][X[:, 0] < 0] = 0
        est2 = sm.OLS(total[:, 1], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                # TODO beta 2 - beta 1 , p values and t values calculation is not correct
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _upload_derived(self, df, table_name):
        print(table_name)
        # print(df)

        df = df.drop(columns='_update_time', errors='ignore')
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

        now: float = time.monotonic()
        print(f'{table_name} costs {now - self._timer}s')
        self._timer = now
        self._updated_count[table_name] += df.shape[0]




import datetime
import traceback

import numpy as np
import pandas as pd
from pandas.tseries.offsets import DateOffset

from ...api.basic import BasicDataApi
from ...view.basic_models import FundRet
from .basic_data_part1 import BasicDataPart1
from .basic_data_helper import BasicDataHelper


class BasicFundRet(BasicDataPart1):

    _DIM_TYPE = {
        'w1_': DateOffset(weeks=1),
        'm1_': DateOffset(months=1),
        'm3_': DateOffset(months=3),
        'm6_': DateOffset(months=6),
        'y1_': DateOffset(years=1),
        'y3_': DateOffset(years=3),
        'y5_': DateOffset(years=5),
    }

    Y_1 = 242
    RISK_FEE_RATE = 0.025

    def __init__(self, data_helper: BasicDataHelper):
        self._data_helper = data_helper
        self.basic_api = BasicDataApi()
        BasicDataPart1.__init__(self, self._data_helper)
        self.ret_res = {}

    def init(self, start_date, end_date):
        self.start_date: datetime.date = start_date if isinstance(start_date, datetime.date) else pd.to_datetime(start_date, infer_datetime_format=True).date()
        self.end_date: datetime.date = end_date if isinstance(end_date, datetime.date) else pd.to_datetime(end_date, infer_datetime_format=True).date()

        # 获取区间内的交易日列表
        dts = BasicDataApi().get_trading_day_list(start_date, end_date).drop(columns='_update_time').datetime
        fund_list = self._fund_info_df.fund_id.tolist()
        self.fund_nav = self.basic_api.get_fund_nav_with_date(self.start_date, self.end_date, fund_list)
        self.fund_nav = self.fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(dts).fillna(method='ffill')
        self.fund_nav_fill = self.fund_nav.fillna(method='bfill').copy()
        _fund_to_enddate_dict = self._fund_info_df[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = _fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:, fund_id] = np.nan
        # 直接使用basic的最新fund scale
        fund_scale = self.basic_api.get_fund_size()
        fund_scale['datetime'] = self.fund_nav.index.array[-1]
        self.ret_res['avg_size'] = fund_scale.pivot_table(index='datetime', columns='fund_id', values='latest_size').iloc[0, :]
        # 基金天数
        self.days = self.fund_nav.count() - 1
        self.date_list = self.fund_nav_fill.index

    def get_annual_ret(self, total_ret):
        '''
        累计年化收益
        exp(log(p[-1] / p[0]) / (trade_days / 242)) - 1
        '''
        log_total_ret = np.log(total_ret)
        ret_yearly = self.days / self.Y_1
        return np.exp(log_total_ret / ret_yearly) - 1

    def get_annual_vol(self):
        '''
        累计年化波动率
        (p.shift(1) / p).std(ddof=1) * np.sqrt((days - 1) / year)
        '''
        diff = self.fund_nav.shift(1) / self.fund_nav
        std = diff.std(ddof=1)
        std_yearly = np.sqrt((self.days - 1) / (self.days / self.Y_1))
        return std * std_yearly

    @staticmethod
    def get_mdd(fund_nav):
        '''
        累计最大回撤
        1 - (p / p.cummax()).min()
        '''
        return 1 - (fund_nav / fund_nav.cummax()).min()

    def calculate_one_date(self):
        # 取最近一天的日期和数据
        last_row = self.fund_nav_fill.iloc[-1, :]
        last_date = self.fund_nav_fill.index.array[-1]
        # 遍历每个需要计算的时间区间
        for name, value in self._DIM_TYPE.items():
            # 根据时间区间计算出begin date(bd)
            bd = (last_date - value).date()
            # 试图寻找bd及其之前的最近一天
            bd_list = self.date_list[self.date_list <= bd]
            if bd_list.empty:
                continue
            # 计算区间内的return
            temp_ret = (last_row / self.fund_nav_fill.loc[bd_list.array[-1], :] - 1)
            self.ret_res[name + 'ret'] = temp_ret

        total_ret = self.fund_nav_fill.iloc[-1] / self.fund_nav_fill.iloc[0]
        self.ret_res['to_date_ret'] = (total_ret - 1)
        self.ret_res['annual_ret'] = self.get_annual_ret(total_ret)
        self.ret_res['vol'] = self.fund_nav.pct_change(1).std(ddof=1)
        annual_vol = self.get_annual_vol()
        self.ret_res['sharpe_ratio'] = ((self.ret_res['annual_ret'] - self.RISK_FEE_RATE) / annual_vol)
        self.ret_res['mdd'] = BasicFundRet.get_mdd(self.fund_nav)
        self.ret_res['recent_y_ret'] = (self.fund_nav_fill.iloc[-1] / self.fund_nav_fill[self.fund_nav_fill.index < datetime.date(self.end_date.year, 1, 1)].iloc[-1] - 1)

        self.result = pd.DataFrame.from_dict(self.ret_res).rename_axis(index='fund_id').reset_index()
        self.result['datetime'] = self.fund_nav.index[-1]
        self.result['info_ratio'] = None  # TODO
        self.result = self.result.replace({np.inf: None, -np.inf: None})
        self.result = self.result[self.result.drop(columns=['fund_id', 'datetime']).notna().any(axis=1)]

    def process_all(self, end_date):
        failed_tasks = []
        try:
            default_begin_date = '20050101'  # 算累计指标用较久的起始日
            self.init(default_begin_date, end_date)
            self.calculate_one_date()
            self._data_helper._upload_basic(self.result, FundRet.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_score')
        return failed_tasks


if __name__ == '__main__':
    bfr = BasicFundRet(BasicDataHelper())
    bfr.process_all('20200903')



import json
import traceback
from multiprocessing import Pool
from ...manager.manager_fund import *
from .derived_data_helper import DerivedDataHelper
from ....resource.data_store import *

class FundManagerFundRankProcessor:

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, dm=None, start_date='20191001', end_date='20200101'):
        # 借用dm处理数据表，不依赖输入时间
        if dm is None:
            self.dm = FundDataManager(start_time=start_date, end_time=end_date, score_manager=FundScoreManager())
            self.dm.init(score_pre_calc=False)
        else:
            self.dm = dm
        self.start_date = self.dm.start_date
        self.end_date = self.dm.end_date
        with BasicDatabaseConnector().managed_session() as quant_session:
            _fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.datetime
                ).filter(
                    FundNav.datetime >= self.start_date - datetime.timedelta(400),
                    FundNav.datetime <= self.end_date,
                )
            _nav_df = pd.read_sql(_fund_nav_query.statement, _fund_nav_query.session.bind)
            self.fund_nav = _nav_df.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value')
            self.fund_nav = self.fund_nav.fillna(method='ffill') + self.fund_nav.fillna(method='bfill') - self.fund_nav.fillna(method='bfill')
            self.fund_ret = self.fund_nav.pct_change(242)
            
        self.datelist = self.dm.dts.trading_days[(self.dm.dts.trading_days.datetime >= self.start_date) & (self.dm.dts.trading_days.datetime <= self.end_date)].datetime
        self.manager_list = self.dm.dts.fund_manager_info.index.unique().tolist()
        self.fund_id_list = self.fund_ret.columns.tolist()

    def loop_item(self, dt):
        try:
            _manager_detail = []
            _mng_info_dt = self.dm.dts.fund_manager_info[(self.dm.dts.fund_manager_info['start_date'] <= dt) & (self.dm.dts.fund_manager_info['end_date'] >= dt)]
            for manager_id in self.manager_list:
                if manager_id not in _mng_info_dt.index:
                    fund_list = []
                else:
                    mng_info_i = _mng_info_dt.loc[manager_id]
                    fund_list = mng_info_i.fund_id
                    fund_list = [fund_list] if isinstance(fund_list, str) else fund_list
                    fund_list = [i for i in self.fund_id_list if i in fund_list]
                    score_dic = self.fund_ret.loc[dt][fund_list].dropna()
                    if not score_dic.empty:
                        score_list = sorted(score_dic.items(), key=lambda x:x[1], reverse=True)
                        fund_list = [_[0] for _ in score_list]
                    else:
                        fund_list = []
                if len(fund_list) > 0:
                    dic = {
                        'mng_id':manager_id,
                        'fund_list':json.dumps(fund_list),
                        'datetime':dt
                    }
                    _manager_detail.append(dic)
            #print(f' finish {dt}')
            return pd.DataFrame(_manager_detail)
        except:
            print(f'boom {dt} failed'  )
            return None

    def calc_history(self):
        date_list = self.datelist
        p = Pool()
        self.result = [i for i in p.imap_unordered(self.loop_item, date_list, 64)]
        self.result = [i for i in self.result if i is not None]
        p.close()
        p.join()
        self.result_df = pd.concat(self.result)
        self._data_helper._upload_derived(self.result_df, FundManagerFundRank.__table__.name)

    def calculate_update(self, dt):
        self.result_df = pd.concat([self.loop_item(dt)])

    def process(self, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()

            start_date = start_date_dt - datetime.timedelta(days = 20) #做data manager 主要用信息表 
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
            self.init(start_date=start_date, end_date=end_date)
            self.calculate_update(end_date_dt)
            self._data_helper._upload_derived(self.result_df, FundManagerFundRank.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('manager_fund_rank')
        return failed_tasks

if __name__ == "__main__":
    dm = DataStore.load_dm()
    self = FundManagerFundRank(DerivedDataHelper())


import pandas as pd
import datetime
from ...wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector
from ...view.raw_models import *
from ...view.basic_models import *
from .derived_data_helper import DerivedDataHelper
from ..raw.raw_data_helper import RawDataHelper
from ..basic.basic_data_helper import BasicDataHelper
from ...api.raw import RawDataApi

class ManagerInfoProcessor:
    '''
    wind基金经理数据比较全 含离职基金经理数据
    东财只含当前基金经理数据
    用wind数据做历史， 用东财数据做更新。对比wind和东财历史数据，解决差异，区分历史和需要更新的数据
    '''

    DEFAULT_DATE = datetime.date(2040,1,1)

    def __init__(self):
        self._raw_api = RawDataApi()
        self.derived_helper = DerivedDataHelper()
        self.basic_helper = BasicDataHelper()

    def init(self):
        #load data
        with RawDatabaseConnector().managed_session() as quant_session:
            _query = quant_session.query(WindFundManager)
            self.wind_manager_df = pd.read_sql(_query.statement, _query.session.bind)

        self.raw_fund_nav: pd.DataFrame = self._raw_api.get_em_fund_nav(columns=('ADJUSTEDNAV',))
        self.raw_fund_nav = self.raw_fund_nav.pivot_table(index='DATES', columns='CODES', values='ADJUSTEDNAV').fillna(method='ffill')

        with BasicDatabaseConnector().managed_session() as quant_session:
            _query = quant_session.query(MngInfo)
            self.manager_df = pd.read_sql(_query.statement, _query.session.bind) 

            _query = quant_session.query(FundInfo)
            self.fund_info = pd.read_sql(_query.statement, _query.session.bind) 

        # 没有基金复用的基金id
        l = self.fund_info.order_book_id.tolist()
        s = list(set(l))
        self.only_1_list = [i for i in s if l.count(i) == 1]

    def wind_fund_nav_begin_date(self, wind_id):
        # 找到基金净值开始日期
        _dt = self.raw_fund_nav[wind_id].dropna().index[0]
        return _dt

    def to_date(self, str_date):
        # 转换wind F_INFO_MANAGER_STARTDATE str ——> datetime
        # 如果为空， 用最新数据填充
        if str_date is not None:
            return datetime.datetime.strptime(str_date, '%Y%m%d').date()
        else:
            return datetime.datetime.today().date()

    def history_data_process(self): 
        # 处理第一步
        #wind F_INFO_MANAGER_STARTDATE 为空，表示公布发布了，但是在获取数据的时刻 2020，6，10左右 基金还为成立，先考虑F_INFO_MANAGER_STARTDATE 不为空的基金数据

        self.w1 =  self.wind_manager_df[~self.wind_manager_df['F_INFO_MANAGER_STARTDATE'].isnull()].copy()
        self.w1['fund_id'] = self.w1.apply(
            lambda x: self.basic_helper._get_fund_id_from_order_book_id(x.F_INFO_WINDCODE.replace('!','.').split('.')[0], self.to_date(x.F_INFO_MANAGER_STARTDATE)), axis=1
        )
        self.w1['F_INFO_MANAGER_STARTDATE'] =  self.w1['F_INFO_MANAGER_STARTDATE'].map(lambda x : datetime.datetime.strptime(x, '%Y%m%d').date() )

        # 东财和万得匹配基金经理名 基金id 和起始日时，起始日会有差别， 人工排除
        ## wind 对的
        wind_right_case = [
            ['杨建华', '200002!0', datetime.date(2009, 8, 24)], # 查过简历 wind对
            ['莫海波', '519193!0', datetime.date(2019, 3, 23)], # 查过基金招募说明 wind 对
            ['刘铭', '004612!0', datetime.date(2018, 11, 10)],  # 查过简历 wind对
            ['刘铭', '004613!0', datetime.date(2018, 11, 10)],  # 查过简历 wind对   
            ['陈薪羽', '007885!0', datetime.date(2020, 1, 4)],  # 查过基金招募说明 wind 对
            ['陈薪羽', '007886!0', datetime.date(2020, 1, 4)],  # 查过基金招募说明 wind 对
            ['吴卫东', '008218!0', datetime.date(2019, 11, 18)], # 查过基金招募说明 wind 对
            ['吴卫东', '008220!0', datetime.date(2019, 11, 18)]  # 查过基金招募说明 wind 对
        ]

        ## 东财对的，基金经理都没有离职情况，只需要改wind起始日就可以
        em_right_case = [
            ['杨谷', '320003!0', datetime.date(2006, 2, 22)], # 查过简历 东财对
            ['王超', '002344!0', datetime.date(2016, 5, 11)], # 查过基金招募说明 东财对
            ['张韵', '002457!0', datetime.date(2016, 3, 1)], # 查过基金招募说明 东财 对
            ['周博洋', '005817!0', datetime.date(2018, 4, 20)],# 查过基金招募说明 东财 对
            ['张航', '005018!0', datetime.date(2019, 9, 16)], # 查过简历 东财对
        ]

        # wind 对华裔人名 不加空格， 东财加
        no_space_name = ['TIAN HUAN']

        # 东财名字 对应wind是中文
        name_replace = {
            'Zhang Phoebe':'张蓓蓓'
        }

        self.no_pair_case = [] # 逐条校验过 85条数据 都是wind没有字条，不是数据不一致导致匹配不上的数据
        for r in self.manager_df.itertuples():
            # 通过 经理名字 基金id 基金经理开始工作时间 3者匹配， 如果1对1， 完美
            # 但是发现两个数据原 经理工作的起始日子错位一天 比如 '徐荔蓉', '450001!0'
            # 匹配逻辑放宽至起始日 在2天以内

            # 东财对于外籍华人 会用拼音加中文表示 比如WEIDONG WU(吴卫东)
            # wind 对某些拼音人名不加 空格 比如 TIAN HUAN
            if '(' in r.mng_name:
                mng_name = r.mng_name.split('(')[1].split(')')[0]
            else:
                mng_name = r.mng_name
                
            if mng_name in no_space_name:
                mng_name = mng_name.replace(' ', '')
            
            if mng_name in name_replace:
                mng_name = name_replace[mng_name]
            
            order_book_id = r.fund_id.replace('!','.').split('.')[0]
            
            if order_book_id not in self.only_1_list:
                continue
            
            # 用东财数据替换wind
            case = [mng_name, r.fund_id, r.start_date]
            if case in em_right_case:
                self.w1.loc[(self.w1['F_INFO_FUNDMANAGER'] == mng_name) \
                            &(self.w1['fund_id'] == r.fund_id),'F_INFO_MANAGER_STARTDATE'] = r.start_date
                continue

            if case in wind_right_case:
                continue
                
            _df = self.w1[(self.w1['F_INFO_FUNDMANAGER'] == mng_name) \
                &(self.w1['fund_id'] == r.fund_id) \
                &(self.w1['F_INFO_MANAGER_STARTDATE'] >= (r.start_date - datetime.timedelta(2))) \
                &(self.w1['F_INFO_MANAGER_STARTDATE'] <= (r.start_date + datetime.timedelta(2)))]
            if _df.shape[0] == 1:
                continue
            else:
                # 没有1对1匹配 放开起始日限制，重新匹配
                em_case = self.manager_df[ (self.manager_df['mng_name'] == mng_name) & \
                    (self.manager_df['fund_id'] == r.fund_id) & \
                    (self.manager_df['start_date'] == r.start_date) ]
            
                wind_case = self.w1[(self.w1['F_INFO_FUNDMANAGER'] == mng_name) \
                    &(self.w1['fund_id'] == r.fund_id)]

                # 1对1 匹配
                if em_case.shape[0] == 1 and wind_case.shape[0] == 1:
                    wind_id = wind_case.F_INFO_WINDCODE.values[0]
                    nav_start_date = self.wind_fund_nav_begin_date(wind_id)
                    
                    # 如果净值开始时间晚于 两个数据基金经理开始时间， 不认为是错误
                    if (nav_start_date >= em_case.start_date.values[0]) and (nav_start_date >= wind_case.F_INFO_MANAGER_STARTDATE.values[0]):
                        continue
                
                self.no_pair_case.append([r.mng_name, r.fund_id, r.start_date])
                
    def history_data_process_dlc(self):
        # 处理第二步
        # 跑完 self.history_data_process 有一些东财未配对的数据
        # 现在处理这些未配对的数据 和 wind 起始时间未空的数据， 合并重复项
        self.w2 =  self.wind_manager_df[self.wind_manager_df['F_INFO_MANAGER_STARTDATE'].isnull()].copy()
        self.w2['fund_id'] = self.w2.apply(
            lambda x: self.basic_helper._get_fund_id_from_order_book_id(x.F_INFO_WINDCODE.replace('!','.').split('.')[0], datetime.date(2020,9,1)), axis=1
        )
        self.new_case = []
        for case_i in self.no_pair_case:
            mng_name = case_i[0]
            fund_id = case_i[1]
            wind_id = fund_id.split('!')[0] +'.OF'
            start_date = case_i[2]
            _df = self.w2[(self.w2['F_INFO_FUNDMANAGER'] == mng_name) & (self.w2['F_INFO_WINDCODE'] == wind_id)]
            # 东财数据新的数据 匹配到wind的历史数据， 更新fund_id 和 基金经理开始任职时间
            if _df.shape[0] == 1:
                self.w2.loc[(self.w2['F_INFO_FUNDMANAGER'] == mng_name) \
                            &(self.w2['F_INFO_WINDCODE'] == wind_id),'F_INFO_MANAGER_STARTDATE'] = start_date
                self.w2.loc[(self.w2['F_INFO_FUNDMANAGER'] == mng_name) \
                            &(self.w2['F_INFO_WINDCODE'] == wind_id),'fund_id'] = fund_id
            else:
                # 找不到到对应数据，船新基金经理数据
                self.new_case.append(case_i)
    
    def existed_mng_id_date_process(self):
        # 处理第三步
        # 如果new_case里数据 基金经理信息 在wind里已有基金id， 用基金id填充
        self.history_data = pd.concat([self.w1,self.w2]).copy()
        self.existed_person_info = []
        self.new_person_info = []
        for case_i in self.new_case:
            mng_name = case_i[0]
            fund_id = case_i[1]
            wind_id = fund_id.split('!')[0] +'.OF'
            start_date = case_i[2]
            mng_info = self.manager_df[(self.manager_df['fund_id'] == fund_id) & (self.manager_df['mng_name'] == mng_name)].copy()
            resume = mng_info.resume.values[0]
            
            # 看是否同名经理 同名人任职过的基金公司名字是否在当前字条的工作经历中 如果是复用基金经理id
            company_now = self.fund_info[self.fund_info['fund_id'] == fund_id].company_id.values[0]
            same_mng_name_fund_ids = fund_list = self.history_data[self.history_data['F_INFO_FUNDMANAGER'] == mng_name].dropna(subset=['fund_id']).fund_id
            if len(same_mng_name_fund_ids) > 0:
                company_list = list(set(self.fund_info[self.fund_info['fund_id'].isin(fund_list)].company_id))
                is_same_person = False
                _fund_id = ''
                for company_i in company_list:
                    if company_i in resume:
                        is_same_person = True
                        company_stayed = company_i
                if is_same_person:
                    stayed_fund_ids = self.fund_info[(self.fund_info['fund_id'].isin(fund_list)) & (self.fund_info['company_id'] == company_stayed )].fund_id.tolist()
                    mng_id = self.history_data[(self.history_data['F_INFO_FUNDMANAGER'] == mng_name) & (self.history_data['fund_id'] == stayed_fund_ids[0])].F_INFO_FUNDMANAGER_ID.values[0]
                    mng_info['F_INFO_FUNDMANAGER_ID'] = mng_id
                    self.existed_person_info.append(mng_info)
                else:
                    self.new_person_info.append(mng_info)

    def make_new_manager_info(self):
        # 处理第四步
        # 不存在基金经理id的人造id
        df = pd.concat(self.new_person_info,axis=0)
        _df = df[['start_mng_date','mng_name']].drop_duplicates().reset_index(drop=True)
        _df['F_INFO_FUNDMANAGER_ID'] = _df.index
        _df['F_INFO_FUNDMANAGER_ID'] = _df['F_INFO_FUNDMANAGER_ID'].map(lambda x: f'mngid_{x+1}')
        mng_id_df = _df.set_index(['start_mng_date','mng_name'])
        df['F_INFO_FUNDMANAGER_ID'] = df.apply(lambda x: mng_id_df.loc[x.start_mng_date, x.mng_name], axis=1)
        self.new_id_mng = df.copy()
        self.old_id_mng = pd.concat(self.existed_person_info,axis=0)
        self.new_data = pd.concat([self.new_id_mng,self.old_id_mng])

    def change_date(self, x):
        if x is not None:
            return datetime.datetime.strptime(x, '%Y%m%d').date()
        return x

    def unify_columns(self):
        # 处理第五步
        # 融合成新数据格式
        self.new_data['end_date'] = None

        col_dict = {
            'F_INFO_FUNDMANAGER_ID':'mng_id',
            'F_INFO_FUNDMANAGER':'mng_name',
            'F_INFO_MANAGER_GENDER':'gender',
            'F_INFO_MANAGER_BIRTHYEAR':'birth_year',
            'F_INFO_MANAGER_EDUCATION':'education',
            'F_INFO_MANAGER_NATIONALITY':'nationality',
            'F_INFO_MANAGER_RESUME':'resume',
            'F_INFO_MANAGER_STARTDATE':'start_date',
            'F_INFO_MANAGER_LEAVEDATE':'end_date',
        }

        wind_columns = ['F_INFO_FUNDMANAGER_ID',
                        'fund_id',
                        'F_INFO_FUNDMANAGER',
                        'F_INFO_MANAGER_GENDER',
                        'F_INFO_MANAGER_BIRTHYEAR',
                        'F_INFO_MANAGER_EDUCATION',
                        'F_INFO_MANAGER_NATIONALITY',
                        'F_INFO_MANAGER_RESUME',
                        'F_INFO_MANAGER_STARTDATE',
                        'F_INFO_MANAGER_LEAVEDATE']
        df_1 = self.history_data[wind_columns].rename(columns=col_dict).copy()
        df_1['gender'] = df_1['gender'].replace('m','男').replace('f','女')
        df_1['end_date'] = df_1['end_date'].map(lambda x: self.change_date(x) )

        em_columns = ['F_INFO_FUNDMANAGER_ID',
                    'fund_id',
                    'mng_name',
                    'gender',
                    'birth_year',
                    'education',
                    'nationality',
                    'resume',
                    'start_date',
                    'end_date']
        df_2 = self.new_data[em_columns].rename(columns=col_dict).copy()
        self.result = pd.concat([df_1,df_2])
        self.result['start_date'] = self.result['start_date'].fillna(self.DEFAULT_DATE)
        self.result['end_date'] = self.result['end_date'].fillna(self.DEFAULT_DATE)
        self.result['fund_id'] = self.result['fund_id'].fillna('not_exist') # TODO update basic.fund_info if too many not_exist existed 

if __name__ == "__main__":
    self = ManagerInfoProcessor()
    self.init()
    self.history_data_process()
    self.history_data_process_dlc()
    self.existed_mng_id_date_process()
    self.make_new_manager_info()
    self.unify_columns()
    self.derived_helper._upload_derived(self.result.drop_duplicates(), 'manager_info')


import pandas as pd
import numpy as np
import datetime
import copy
from pprint import pprint 
import platform
import matplotlib as mpl
import json
from collections import Counter
from .painter.asset_painter import AssetPainter
from .painter.fund_painter import FundPainter
from .painter.hybrid_painter import HybridPainter
from ...data.struct import AssetWeight, AssetPosition, AssetPrice, AssetValue
from .asset_helper import TAAHelper
from ...data.manager.manager_fund import FundDataManager
from ...data.struct import FundTrade, TaaTunerParam, FundPosition, TAAParam
from .asset_helper import SAAHelper, TAAHelper, FAHelper, TAAStatusMode
from ...util.calculator import Calculator

CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']

class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(DatetimeEncoder, obj).default(obj)
        except TypeError:
            return str(obj)

class ReportHelper:
    
    '''
    save backtest history
    '''

    def __init__(self):
        pass

    def init(self): 
        pass

    def plot_init(self, dm, taa_helper=None, score_select=None):
        self.index_price = dm.get_index_price()
        self.fund_info = dm.get_fund_info()
        self.fund_nav = dm.dts.fund_nav
        self.fund_indicator = dm.dts.fund_indicator
        self.taa_params = taa_helper.params if isinstance(taa_helper, TAAHelper) else None
        self.index_pct = dm.dts.index_pct
        self.dts = dm.dts.trading_days.datetime
        self.fund_score_dm = dm._score_manager.score_cache
        trade_list = []
        for d, trade_d in self.trade_history.items():
            trade_d = [_.__dict__ for _ in trade_d]
            trade_list.extend(trade_d)
        self.trade_df = pd.DataFrame(trade_list)
        if not self.trade_df.empty:
            self.trade_df['year'] = [_.year for _ in self.trade_df.trade_date]
        self.pct_df = dm.dts._index_pct_df
        self.score_select = score_select
        self.active_fund_info = dm.dts.active_fund_info
        self.fund_info = self.fund_info.set_index('fund_id')
        for r in self.active_fund_info.itertuples():
            self.fund_info.loc[r.fund_id,'index_id'] = r.index_id
        self.fund_info = self.fund_info.reset_index()

    def setup(self, saa:AssetWeight=None):
        self.saa_weight = saa.__dict__ if saa else None
        self.asset_cash_history = {}
        self.asset_position_history = {}
        self.asset_market_price_history = {}
        self.pending_trade_history = {}
        self.asset_weights_history = {}
        self.tactic_history = {}
        self.fund_position_history = {}
        self.trade_history = {}
        self.rebalance_date = []
        self.fund_cash_history = {}
        self.fund_marktet_price_history = {}
        self.fund_weights_history = {}
        self.fund_score = {}
        self.fund_score_raw = {}
        self.target_allocation = {}
        self._rebalance_details = []

    def update(self,dt:datetime, asset_cur_position:AssetPosition, asset_cur_price:AssetPrice, pend_trades:list, fund_cur_position:FundPosition, fund_nav:dict, traded_list:list, fund_score:dict, fund_score_raw:dict, target_allocation):   
        # 检查回测时使用
        self.asset_cash_history[dt] = asset_cur_position.cash##
        self.asset_position_history[dt] = asset_cur_position.__dict__##
        self.pending_trade_history[dt] = pend_trades##
        if fund_cur_position is not None:
            dic = {f : f_info.__dict__  for f, f_info in fund_cur_position.copy().funds.items()}
            for f, f_info in dic.items():
                f_info['price'] =  fund_nav[f]
            self.fund_position_history[dt] = dic     
            self.fund_cash_history[dt] = fund_cur_position.cash
            mv, fund_w = fund_cur_position.calc_mv_n_w(fund_navs=fund_nav)
            self.fund_marktet_price_history[dt] = mv
            self.fund_weights_history[dt] = { fund_id : w_i for fund_id, w_i in fund_w.items() if w_i > 0}
            self.fund_score[dt] = fund_score
            self.fund_score_raw[dt] = fund_score_raw

        self.asset_market_price_history[dt] = AssetValue(prices=asset_cur_price, positions=asset_cur_position).sum() 
        asset_w = AssetValue(prices=asset_cur_price, positions=asset_cur_position).get_weight().__dict__
        self.asset_weights_history[dt] = { index_id : w_i for index_id, w_i in asset_w.items() if w_i > 0}
        if traded_list is not None:
            if len(traded_list) > 0:
                self.rebalance_date.append(dt)
                self.trade_history[dt] = traded_list
        self.target_allocation[dt] = target_allocation
        
    def _calc_stat_yearly(self, mv_df):
        if not self.trade_df.empty:
            trade_df = self.trade_df.set_index(['year','trade_date'])
            trade_df = trade_df[trade_df['is_buy'] == True]
            trade_amount_year = trade_df.groupby('year').sum()
            mv_df = mv_df.reset_index()
            dates = mv_df['date'].values
            mv_df['year'] = [_.year for _ in mv_df.date]
            mv_df = mv_df.set_index(['year','date'])
            res = []
            for year in trade_amount_year.index:
                mv_year = mv_df.loc[year]
                dic = {
                    'year': year,
                    'yearly_amount': trade_amount_year.loc[year, 'amount'],
                    'year_begin_mv': mv_df.loc[year].iloc[0].mv,
                    'year_end_mv':mv_year.iloc[-1].mv,
                    'year_mdd':1 - (mv_year.mv / mv_year.mv.cummax()).min()
                    }
                res.append(dic)
            self.turnover_df = pd.DataFrame(res)
            self.turnover_df['turnover_rate_yearly'] = 100 * self.turnover_df['yearly_amount'] / self.turnover_df['year_begin_mv']
            self.turnover_df['year_ret'] = self.turnover_df.year_end_mv / self.turnover_df.year_begin_mv - 1
            self.turnover_rate_yearly_avg = self.turnover_df['turnover_rate_yearly'].sum() / (mv_df.index[-1][0] - mv_df.index[0][0] + 1)
            total_amount = self.trade_df[self.trade_df['is_buy'] == True].amount.sum()
            self.turnover_rate_amt_mv_gmean = Calculator.get_turnover_rate(dates=dates, values=mv_df.mv.values, total_amount=total_amount)
        else:
            self.turnover_df = pd.DataFrame()
            self.turnover_rate_yearly_avg = 0
            self.turnover_rate_amt_mv_gmean = 0

    def _calc_stat(self, df):
        res = Calculator.get_stat_result_from_df(df=df.reset_index(), date_column='date', value_column='mv')
        w = copy.deepcopy(self.saa_weight) if self.saa_weight else AssetWeight().__dict__ # asset weight float
        self._calc_stat_yearly(df)
        w['mdd'] = res.mdd #float
        w['annual_ret'] = res.annualized_ret #float
        w['ret_over_mdd'] = res.ret_over_mdd #float
        w['sharpe'] = res.sharpe #float 
        w['5_year_ret'] = res.recent_y5_ret #float
        w['annual_vol'] = res.annualized_vol #float
        w['mdd_d1'] = res.mdd_date1 #datetime
        w['mdd_d2'] = res.mdd_date2 #datetime
        w['turnover_rate_yearly_avg'] = self.turnover_rate_yearly_avg #float 两种算法的换手率结果很接近
        w['turnover_rate_amt_mv_gmean'] = self.turnover_rate_amt_mv_gmean #float
        w['start_date'] = res.start_date #datetime
        w['end_date'] = res.end_date #datetime
        if not self.trade_df.empty:
            w['total_fee_over_begin_mv'] = self.trade_df.commission.sum() / df.mv[0] #float
            w['total_commission'] = self.trade_df.commission.sum()
        else:
             w['total_fee_over_begin_mv'] = 0
             w['total_commission'] = 0
        w['rebalance_date'] = self.rebalance_date #list
        w['last_unit_nav'] = res.last_unit_nav
        w['last_mv_diff'] = res.last_mv_diff
        w['last_increase_rate'] = res.last_increase_rate
        w['market_value'] = df #dataframe
        w['rebalance_times'] = len(self.rebalance_date)
        w['recent_drawdown'] = res.recent_drawdown
        w['worst_3m_ret'] = res.worst_3m_ret
        w['worst_6m_ret'] = res.worst_6m_ret
        hold_days = self.rebalance_date + [df.index.values[-1]]
        index_list = df.index.tolist()
        result = {}
        for d in hold_days[:-1]:
            d1 = hold_days[hold_days.index(d) + 1]
            idx0 = index_list.index(d)
            idx1 = index_list.index(d1)
            result[d.strftime('%Y-%m-%d')] = (idx1 - idx0) / 242
        w['hold_years'] = result # dic
        return w

    def get_asset_stat(self):
        self.asset_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.asset_market_price_history.items()]).set_index('date')
        w = self._calc_stat(self.asset_mv.copy())
        return w
    
    def save_asset_bk(self, json_name):
        dic = self.get_asset_stat()
        dic['market_value'] =  dic['market_value'].reset_index().to_dict('records')
        with open(f'{json_name}.json','w') as f:
            f.write(json.dumps(dic, cls=DatetimeEncoder))

    def get_fund_stat(self):
        if not self.fund_marktet_price_history:
            return None
            
        self.fund_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.fund_marktet_price_history.items()]).set_index('date')
        w = self._calc_stat(self.fund_mv.copy())
        return w

    def get_last_position(self):
        last_date = list(self.fund_position_history.keys())[-1]
        dic = self.fund_position_history[last_date]
        weight_dict = self.fund_weights_history[last_date]
        name_dict =  self.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        dic_list = [v for k,v in dic.items() if v['volume'] > 0]
        pos_df = pd.DataFrame(dic_list).sort_values('index_id')[['fund_id','index_id']].set_index('fund_id')
        pos_df['weight'] = pos_df.index.map(lambda x : weight_dict[x])
        pos_df['desc_name'] = pos_df.index.map(lambda x : name_dict[x])
        pos_df.index.name = last_date
        pos_df = pos_df.sort_values('weight', ascending = False)
        dt = sorted(list(self.fund_score_dm.keys()))[-1]
        fund_rank = []
        for i in pos_df.itertuples():
            _score_dict = self.fund_score_dm[dt][i.index_id][self.score_select[i.index_id]]
            _score_list = sorted(_score_dict.items(), key=lambda x:x[1], reverse=True)
            _fund_list = [i[0] for i in _score_list]
            rank_i = _fund_list.index(i.Index) + 1 if i.Index in _fund_list else 10000
            fund_rank.append(rank_i)
        pos_df.loc[:,'fund_rank'] = fund_rank
        return pos_df

    def get_last_target_fund_allocation(self, target_fund):
        res = []
        fund_info_dic = self.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        for fund_id, fund_w in target_fund.items():
            dic = {'fund_id':fund_w.fund_id,
                'index_id':fund_w.index_id, 
                'weight':round(fund_w.fund_wgt,3),
                'desc_name':fund_info_dic[fund_w.fund_id]}
            res.append(dic)
        target_fund = pd.DataFrame(res).set_index('fund_id')
        target_fund.index.name = list(self.fund_position_history.keys())[-1]
        return target_fund.sort_values('weight', ascending = False)[['index_id','weight','desc_name']]

    def get_fund_trade(self):
        fsdf = self.fund_score.copy()
        fund_info_df = self.fund_info.copy().set_index('fund_id')
        fund_trade = []
        for d in self.trade_history:
            f_t = [ i.__dict__  for i in self.trade_history[d] if isinstance(i, FundTrade)]
            fund_trade.extend(f_t)
        if len(fund_trade) < 1:
            return pd.DataFrame()
        ft_res = []
        for ft in fund_trade:
            ft['desc_name'] = fund_info_df.loc[ft['fund_id'], 'desc_name']
            index_id = fund_info_df.loc[ft['fund_id'], 'index_id']
            try:
                s = fsdf[ft['submit_date']][index_id][ft['fund_id']]
            except:
                s = np.nan
            ft['submit_d_score'] = s
            fund_id = ft['fund_id']
            submit_d = ft['submit_date']
            traded_d = ft['trade_date']
            ft['before_w']  = self.fund_weights_history[submit_d].get(fund_id,0)
            ft['after_w'] = self.fund_weights_history[traded_d].get(fund_id,0)
            ft_res.append(ft)
        df = pd.DataFrame(ft_res)
        df['before_w'] = df['before_w'].round(4)
        df['after_w'] = df['after_w'].round(4)
        dts = df.submit_date.unique().tolist()
        for dt in dts:
            _df = df[df['submit_date'] == dt]
            fund_id = _df.iloc[-1].fund_id
            df.loc[(df['submit_date'] == dt) & (df['fund_id'] == fund_id),'after_w'] = 1 - _df.iloc[:-1].after_w.sum()
            df.loc[(df['submit_date'] == dt) & (df['fund_id'] == fund_id),'before_w'] = 1 - _df.iloc[:-1].before_w.sum()
        return df

    @staticmethod
    def fund_rank_recent_analysis(dm, fund_id, score_select):
        score_dict = dm._score_manager.score_cache
        fund_info = dm.dts.fund_info
        fund_indicator = dm.dts.fund_indicator
        fund_nav = dm.dts.fund_nav
        index_price = dm.dts.index_price
        index_id = fund_info[fund_info['fund_id'] == fund_id].index_id.values[0]
        score_func = dm._score_manager.funcs[index_id][score_select[index_id]]
        return HybridPainter.fund_rank_recent_analysis(fund_id,index_id,score_dict,fund_info,fund_indicator,fund_nav,index_price,score_func,score_select)

    def get_asset_trade(self):
        asset_trade = []
        for d in self.trade_history:
            a_t = [ i.__dict__  for i in self.trade_history[d] if not isinstance(i, FundTrade)]
            asset_trade.extend(a_t) 
        res = []
        for dic in asset_trade: 
            index_id = dic['index_id']
            submit_d = dic['submit_date']
            traded_d = dic['trade_date']
            dic['before_w'] = self.asset_weights_history[submit_d].get(index_id, 0)
            dic['after_w'] = self.asset_weights_history[traded_d].get(index_id, 0)
            res.append(dic)
        df = pd.DataFrame(res)
        df['before_w'] = df['before_w'].round(4)
        df['after_w'] = df['after_w'].round(4)
        dts = df.submit_date.unique().tolist()
        for dt in dts:
            _df = df[df['submit_date'] == dt]
            index_id = _df.iloc[-1].index_id
            df.loc[(df['submit_date'] == dt) & (df['index_id'] == index_id),'after_w'] = 1 - _df.iloc[:-1].after_w.sum()
            df.loc[(df['submit_date'] == dt) & (df['index_id'] == index_id),'before_w'] = 1 - _df.iloc[:-1].before_w.sum()
        return df

    def get_fund_position(self):
        fund_pos = []
        for d in self.fund_position_history:
            for i in self.fund_position_history[d].values():
                if i['volume'] > 0:
                    i['datetime'] = d
                    fund_pos.append(i)
        return pd.DataFrame(fund_pos).sort_values('datetime')

    def get_index_fee_avg(self):
        self.index_fee = self.trade_df.copy()
        if self.index_fee.empty:
            self.index_fee_buy = pd.DataFrame()
            self.index_fee_sel = pd.DataFrame()
            self.index_fee_buy['fee_ratio_avg'] = 0
            self.index_fee_sel['fee_ratio_avg'] = 0
            self.index_fee['fee_ratio_avg'] = 0
        else:
            self.index_fee_buy = self.index_fee[self.index_fee.is_buy].groupby('index_id').sum()
            self.index_fee_sel = self.index_fee[np.logical_not(self.index_fee.is_buy)].groupby('index_id').sum()
            self.index_fee_buy['fee_ratio_avg'] = self.index_fee_buy.commission / self.index_fee_buy.amount
            self.index_fee_sel['fee_ratio_avg'] = self.index_fee_sel.commission / self.index_fee_sel.amount
            self.index_fee = self.index_fee.groupby('index_id').sum()
            self.index_fee['fee_ratio_avg'] = self.index_fee.commission / self.index_fee.amount

    def get_rebalance_detail(self):
        #rebalance_details = [i['rebalance_reason'][0]['rebalanace_reason'] for i in self._rebalance_details]
        #rebalance_details = dict(Counter(rebalance_details))
        trade_df = self.trade_df.copy()
        fund_mv = self.fund_mv.copy()
        result = self._calc_stat(fund_mv).copy()
        _rb_df = pd.DataFrame(self._rebalance_details)
        submit_date_to_trade_day = trade_df[['submit_date','trade_date']].set_index('submit_date').to_dict()['trade_date']
        turnover_df = trade_df[trade_df['is_buy'] == True].groupby('submit_date').sum()[['amount']].join(fund_mv).copy().reset_index()
        turnover_df['trade_date'] = turnover_df['submit_date'].map(lambda x : submit_date_to_trade_day[x])
        turnover_df = turnover_df[['amount','mv','trade_date']].set_index('trade_date')
        _rb_df['trade_date'] = _rb_df['datetime'].map(lambda x : submit_date_to_trade_day[x])
        _rb_df['rebalance reason'] = _rb_df['rebalance_reason'].map(lambda x : x[0]['rebalanace_reason'])
        _res = []
        for dic in _rb_df['rebalance_reason']:
            dic = dic[0]
            strs = [ f'{v} ' for k , v in dic.items() if k is not 'rebalanace_reason']
            strs = ''.join(strs)
            _res.append(strs)
        _rb_df['rebalance_detail'] = _res
        _rb_df = _rb_df[['trade_date','rebalance reason','rebalance_detail']].set_index('trade_date').copy()
        turnover_df['turnover'] = turnover_df['amount'] / turnover_df['mv']
        _df = pd.DataFrame([{'trade_date':d,'hold_year':l} for d,l in result['hold_years'].items()]).set_index('trade_date')
        _df.index = pd.to_datetime(_df.index)
        turnover_df = turnover_df.join(_df).join(_rb_df).copy()
        return turnover_df[['turnover','hold_year','rebalance reason','rebalance_detail']]

    def backtest_asset_plot(self):
        bt_type = 'asset'
        print(f'{bt_type} report')
        w = self.get_asset_stat()
        del w['market_value']
        pprint(w)
        AssetPainter.plot_asset_weights(self.asset_weights_history, self.saa_weight)
        HybridPainter.plot_market_value(self.asset_mv, bt_type, self.index_price, self.saa_weight, w)
        AssetPainter.plot_asset_mdd_period(self.asset_mv, self.saa_weight, self.index_price, self.asset_weights_history)
        AssetPainter.plot_asset_mdd_amount(self.asset_mv, self.index_price, self.asset_position_history)

    def backtest_fund_plot(self):
        bt_type = 'fund'
        print(f'{bt_type} report')
        w = self.get_fund_stat()
        mv = w['market_value'].copy()
        del w['market_value']
        pprint(w)
        w['market_value'] = mv
        self.get_index_fee_avg()
        #FundPainter.plot_fund_weights(self.fund_weights_history, self.fund_cash_history, self.fund_marktet_price_history, self.saa_weight)
        HybridPainter.plot_market_value_with_mdd_analysis(self.fund_mv, bt_type, self.index_price, self.saa_weight, w)
        FundPainter.plot_fund_mdd_periods(self.fund_mv, self.fund_weights_history, self.fund_nav, self.fund_info)
        FundPainter.plot_fund_mdd_amounts(self.fund_mv, self.fund_nav, self.fund_info, self.fund_position_history.copy())
        FundPainter.plot_fund_alpha(self.fund_nav, self.fund_info, self.index_price, self.fund_position_history.copy(), w)
        #FundPainter.plot_mv_on_each_asset(self.fund_nav, self.fund_info, self.index_price, self.fund_position_history.copy(), w)
        #HybridPainter.plot_asset_fund_mv_diff(self.asset_mv, self.fund_mv)
        # if not self.index_fee.empty:
        #     FundPainter.plot_index_fund_fee(self.index_fee)
        # if not self.turnover_df.empty:
        #     FundPainter.plot_fund_ret_each_year(self.turnover_df, w['mdd'], w['annual_ret'] )
        #     FundPainter.plot_turnover_rate_each_year(self.turnover_df, self.turnover_rate_yearly_avg)

    def singel_index_plot(self):
        w = self.get_fund_stat()
        HybridPainter.plot_market_value_with_mdd_analysis(self.fund_mv, 'fund', self.index_price, self.saa_weight, w)
        FundPainter.plot_fund_mdd_periods(self.fund_mv, self.fund_weights_history, self.fund_nav, self.fund_info)
        FundPainter.plot_fund_alpha(self.fund_nav, self.fund_info, self.index_price, self.fund_position_history.copy(), w)

    def asset_price_ratio(self):    
        fund_trade = self.get_fund_trade()
        return HybridPainter.plot_asset_price_ratio(self.fund_nav, self.fund_info, self.index_price, self.fund_score_dm, self.get_fund_stat(), self.dts, self.saa_weight, fund_trade, self.score_select)
        
    def plot_whole(self):
        self.backtest_fund_plot()
        self.backtest_asset_plot()
        
    def _plot_fund_score(self, asset, is_tuning):
        FundPainter.plot_fund_score(self.fund_mv, 
                                    self.fund_weights_history, 
                                    self.trade_history,
                                    self.index_price, 
                                    self.asset_weights_history,
                                    self.fund_info,
                                    self.fund_nav,
                                    self.fund_score,
                                    self.fund_score_raw,
                                    self.fund_indicator,
                                    asset,
                                    is_tuning,
                                    self.score_select,
                                    )
   
    def _index_pct_plot(self, index_id:str, saa_mv:pd.DataFrame, taa_mv:pd.DataFrame):
        pct = TaaTunerParam.POOL[index_id]
        index_pct = self.index_pct.xs(index_id, level=1, drop_level=True)[[pct]].rename(columns={pct:index_id})
        AssetPainter.plot_taa_analysis(saa_mv, taa_mv, index_id, index_pct, self.taa_params, self.index_price)
        
    def _plot_taa_saa(self, saa_mv, taa_mv, index_id):
        pct = TaaTunerParam.POOL[index_id]
        index_pct = self.index_pct.xs(index_id, level=1, drop_level=True)[[pct]].rename(columns={pct:index_id})
        AssetPainter.plot_asset_taa_saa(saa_mv, taa_mv, index_id, index_pct)
       
    def save_fund_bk_data(self, csv_title:str='check'):
        self.fund_mv.to_csv(csv_title + '_fund_mv.csv')
        self.get_fund_trade().to_csv(csv_title + '_fund_trade.csv')
        self.get_fund_position().to_csv(csv_title + '_fund_pos.csv')
        pd.DataFrame(self.asset_position_history).T.to_csv(csv_title + '_asset_pos.csv')
        if hasattr(self, 'asset_mv'):
            self.asset_mv.to_csv(csv_title + '_asset_mv.csv')
        self.get_asset_trade().to_csv(csv_title + '_asset_trade.csv')
    
    def update_rebalance_detail(self, dt, rebalance_reason):
        if bool(rebalance_reason):
            self._rebalance_details.append({'datetime':dt,'rebalance_reason':rebalance_reason})

    def plot_taa_pct_df(self, dt):
        res = []
        for index_id, pct_data in TaaTunerParam.POOL.items():
            res.append( self.pct_df.loc[index_id][['datetime',pct_data]].reset_index().pivot_table(index='datetime', columns='index_id')[pct_data])
        pct_df = pd.concat(res, axis=1, sort=True)
        df = pct_df.loc[dt:]
        HybridPainter.plot_taa_detail(df)

    @staticmethod
    def get_taa_result(index_id:str, start_date:datetime.date, end_date:datetime.date, taaParam:TAAParam, dm:FundDataManager):
        taa_helper = TAAHelper(taa_params=taaParam)
        _dts = dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime 
        fake_weight = AssetWeight(**{index_id:1})
        for dt in dts:
            asset_pct = dm.get_index_pcts_df(dt)
            taa_helper.on_price(dt=dt, asset_price=None, cur_saa=fake_weight, asset_pct=asset_pct, select_val={},score_dict={})
        df = pd.DataFrame(taa_helper.tactic_history).T.dropna()
        df['last_date'] = df.index.to_series().shift(1)
        con = df[index_id] != df[index_id].shift(1)
        df_diff_part = df[con].copy()
        df_diff_part = df_diff_part.reset_index().rename(columns={'index':'begin_date'})
        next_date = df_diff_part['last_date'].shift(-1).tolist()
        next_date[-1] = df.index.values[-1]
        df_diff_part['end_date'] = next_date
        df_result = df_diff_part[df_diff_part[index_id] != TAAStatusMode.NORMAL][['begin_date','end_date',index_id]]
        df_result = df_result.rename(columns = {index_id:'status'})
        return df_result.to_dict('records') 

    def get_date_score(self, datetime:datetime.date):
        fund_score = self.fund_score[datetime]
        fund_score_raw = self.fund_score_raw[datetime]
        desc_name_dic = self.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        res = []
        for index_id, score_dic in fund_score.items():
            for fund_id, fund_s in score_dic.items():
                dic = {
                    'index_id' : index_id,
                    'fund_id': fund_id,
                    'desc_name':desc_name_dic[fund_id],
                    'score':fund_s,
                    'raw_score':fund_score_raw[index_id][fund_id],
                }
                res.append(dic)
        return pd.DataFrame(res).sort_values(['index_id','score'],ascending=[True,False]).set_index('index_id')

    def get_mdd_recover_stats(self):
        fund_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.fund_marktet_price_history.items()]).set_index('date')
        res = Calculator.get_mdd_recover_result(fund_mv)
        return res

    def get_mdd_stats(self):
        fund_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.fund_marktet_price_history.items()]).set_index('date')
        res = Calculator.get_mdd_result(fund_mv)
        return res

    @staticmethod
    def plot_fund_score(fund_id:str='200002!0',begin_date:datetime.date=datetime.date(2015,1,1),dm:FundDataManager=None):
        FundPainter.plot_fund_rank(score_dic=dm._score_manager.score_cache,
                                   fund_info=dm.dts.fund_info,
                                   fund_id=fund_id,
                                   begin_date=begin_date,
                                   fund_index_map=dm.dts.fund_index_map)

    @staticmethod
    def indicator_score_d(begin_d:datetime.date, fund_list:list, direction:str, m:FundDataManager):
        indicator_1 = m.dts.fund_indicator[(m.dts.fund_indicator.fund_id.isin(fund_list)) &
                            (m.dts.fund_indicator.datetime == begin_d)]
        indicator_1 = indicator_1[['fund_id','alpha','beta','track_err','fee_rate','index_id']].rename(columns={'alpha':f'{direction}_alpha',
                                                                                                            'beta':f'{direction}_beta',
                                                                                                            'track_err':f'{direction}_track_err'})
        indicator_1 = indicator_1.set_index('fund_id')
        score_day_dic = m._score_manager.score_cache[begin_d]
        fund_ranks = {}
        for index_id, score_dic in score_day_dic.items():
            _rank = sorted(score_dic.items(), key=lambda x:x[1], reverse=True)
            fund_rank = [i[0] for i in _rank]
            fund_ranks[index_id] = fund_rank
        rank_result = []
        for r in indicator_1.itertuples():
            try:
                rank_result.append(fund_ranks[r.index_id].index(r.Index) + 1)   
            except:
                rank_result.append(10000)
        indicator_1.loc[:,f'{direction}_rank'] = rank_result
        return indicator_1
     
    @staticmethod
    def fund_trade_result(_bk_class,_bk_result,m):
        _dts = m.dts.trading_days.datetime.tolist()
        rebalance_date = _bk_result['rebalance_date']
        begin_date = rebalance_date[0]
        mv_df = _bk_result['market_value']
        rebalance_list = [begin_date] + [i for i in rebalance_date if i > begin_date] + [mv_df.index[-1]]
        fund_info_dic = m.dts.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        df_list = []
        for i in range(len(rebalance_list)-1):
            begin_d = rebalance_list[i]
            end_d = rebalance_list[i+1]
            score_day_1 = _dts[_dts.index(begin_d) -1]
            score_day_2 = _dts[_dts.index(end_d) -1]
            pos_dic = _bk_class._report_helper.fund_position_history[begin_d]
            res = []
            for fund_id, fund_pos_i in pos_dic.items():
                if fund_pos_i['volume'] > 1:
                    res.append({
                        'fund_id':fund_pos_i['fund_id'],
                        'volume':fund_pos_i['volume'],
                        'price1':fund_pos_i['price'],
                        'index_id':fund_pos_i['index_id'],
                    })
            pos_i_df = pd.DataFrame(res)
            pos_i_df.loc[:,'price2'] = pos_i_df['fund_id'].map(lambda x : m.dts.fund_nav.loc[end_d,x])
            pos_i_df['amount'] = pos_i_df['volume'] * pos_i_df['price1']
            pos_i_df['weight'] = pos_i_df['amount'] / pos_i_df['amount'].sum()
            pos_i_df.loc[:,'desc_name'] = pos_i_df['fund_id'].map(lambda x: fund_info_dic[x])
            pos_i_df['fund_ret_rate'] = (pos_i_df['price2'] / pos_i_df['price1']) - 1
            pos_i_df['fund_return'] = pos_i_df['fund_ret_rate'] * pos_i_df['weight'] 
            pos_i_df_show = pos_i_df[['fund_id','desc_name','index_id','weight','fund_ret_rate','fund_return']]
            pos_i_df_show = pos_i_df_show.sort_values('fund_return', ascending=False)
            fund_list = pos_i_df_show.fund_id.tolist()
            pos_i_df_show = pos_i_df_show.set_index('fund_id')
            indicator_1 = ReportHelper.indicator_score_d(score_day_1, fund_list, 'buy', m)
            indicator_2 = ReportHelper.indicator_score_d(score_day_2, fund_list, 'sell', m)
            pos_i_df_show = pos_i_df_show.join(indicator_1[['buy_alpha','buy_beta','buy_track_err','fee_rate','buy_rank']]).join(indicator_2[['sell_alpha','sell_beta','sell_track_err','sell_rank']])
            pos_i_df_show = pos_i_df_show.reset_index()
            pos_i_df_show.index.name = f'{begin_d}~{end_d}'
            pos_i_df_show.loc[:,'buy_signal_date'] = begin_d
            pos_i_df_show.loc[:,'sell_signal_date'] = end_d
            df_list.append(pos_i_df_show)
        return pd.concat(df_list)

    @staticmethod
    def recent_week_return(pos_df,fund_result,weight_df,saa,mv_df,m,benchmark,end_date=None,begin_date=None,port_id='',fund_info=None):
        # 把买卖交易费平滑打入净值中
        def _index_price_ratio_item(index_ratio_res, d1, d2, is_buy_fee, is_sell_fee):
            _index_ratio_res = []
            for index_id, fund_list in index_ratio_res.items():
                _nav = m.dts.fund_nav[fund_list].loc[d1:d2]
                if is_buy_fee:
                    for fund_id in _nav:
                        fund_i_fee = buy_dic.get(fund_id,0)
                        fee_discount_l = np.linspace(1,1-fund_i_fee,_nav.shape[0])
                        _nav[fund_id] = _nav[fund_id] * fee_discount_l
                if is_sell_fee:
                    for fund_id in _nav:
                        fund_i_fee = sell_dic.get(fund_id,0)
                        fee_discount_l = np.linspace(1,1-fund_i_fee,_nav.shape[0])
                        _nav[fund_id] = _nav[fund_id] * fee_discount_l
                _nav['index_price'] = m.dts.index_price.loc[d1:d2,index_id]
                _nav = _nav/_nav.iloc[0]
                _nav['mean'] = _nav.mean(axis=1)
                _nav.loc[:,index_id] = _nav['mean'] / _nav['index_price']
                _index_ratio_res.append(_nav[[index_id]].copy())   
            res_df = pd.concat(_index_ratio_res, sort=True, axis=1)
            return res_df

        def _ret_item(begin_date, today, is_buy_fee, is_sell_fee):
            _fund_list = list(weight_df[begin_date].keys())
            _nav = m.dts.fund_nav[_fund_list].loc[begin_date:today]
            fund_ret = _nav.iloc[-1] / _nav.iloc[0] - 1
            ret_df = pd.DataFrame(fund_ret).rename(columns={0:'ret'})
            res = []
            for k,w in weight_df[begin_date].items():
                res.append({'fund_id':k,'weight':w})
            _weight_df = pd.DataFrame(res).set_index('fund_id')
            res_df = ret_df.join(_weight_df)
            res_df.loc[:,'rets'] = res_df['ret'] * res_df['weight']
            res_df.loc[:,'begin_date'] = begin_date
            res_df.loc[:,'end_date'] = today
            return res_df

        buy_dic = (fund_info[['fund_id','purchase_fee']].set_index('fund_id') * 0.15).to_dict()['purchase_fee']
        sell_dic = fund_info[['fund_id','redeem_fee']].set_index('fund_id').to_dict()['redeem_fee']
        index_price = m.dts.index_price
        if end_date is None:
            today = datetime.date.today()
        else:
            today = end_date
        if begin_date is None:
            begin_date =  today - datetime.timedelta(days=today.weekday())
        trading_days = sorted(list(set(m.dts.trading_days.datetime) & set(mv_df.index)))
        mv_df = mv_df.loc[begin_date:].reindex(trading_days)
        stats = Calculator.get_stat_result_from_df(df=mv_df.dropna().reset_index(), date_column='date', value_column='mv')
        annual_ret= round(stats.annualized_ret,3)
        mdd = round(stats.mdd,4)
        df_plot_i = mv_df.join(m.dts.index_price[[benchmark]]).loc[begin_date:today]
        df_plot_i = df_plot_i / df_plot_i.iloc[0]
        total_ret = int((stats.last_unit_nav - 1) * 10000)
        st = f'check period annual ret : {annual_ret} mdd : {mdd} total_ret {total_ret} bp'
        FundPainter.plot_recent_ret(df_plot_i,begin_date,today,st,port_id)
        _index_list = []
        for asset, w in saa.__dict__.items():
            if w > 0 and asset != 'cash':
                _index_list.append(asset)
        _index_price = index_price.loc[begin_date:today]
        _index_price = _index_price / _index_price.iloc[0]
        AssetPainter.plot_index_price_ratio(_index_price,'index_price')
        
        if begin_date >= fund_result['rebalance_date'][-1]:
            res_df = _ret_item(begin_date,today, False, False)
        else:
            date_list = [[begin_date,fund_result['rebalance_date'][-1]],[fund_result['rebalance_date'][-1],today]]
            _res_dt_df = []
            for idx,dl in enumerate(date_list):
                b1 = dl[0]
                b2 = dl[1]
                _res_dt_df.append(_ret_item(b1, b2, False, False))
            res_df = pd.concat(_res_dt_df, axis=0, sort=True)
        #res_df.loc[:,'rets'] = res_df['ret'] * res_df['weight']
        res_df = res_df.sort_values('rets',ascending=False)
        res_df.loc[:,'rets'] = [str(int(i*10000))+'BP' for i in res_df.loc[:,'rets']]
        desc_dic = fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        index_dic = fund_info[['fund_id','index_id']].set_index('fund_id').to_dict()['index_id']
        res_df['desc_name'] = res_df.index.map(lambda x : desc_dic[x])
        res_df['index_id'] = res_df.index.map(lambda x : index_dic[x])        
        if begin_date > fund_result['rebalance_date'][-1]:
            d1 = begin_date
            d2 = today
            index_ratio_res = {}
            _fund_list = list(weight_df[begin_date].keys())
            for fund_id in _fund_list:
                _index_id = index_dic[fund_id]
                if _index_id not in index_ratio_res:
                    index_ratio_res[_index_id] = [fund_id]
                else:
                    index_ratio_res[_index_id].append(fund_id)
            res_index_df = _index_price_ratio_item(index_ratio_res, d1, d2, False, False)
        else:
            date_list = [[begin_date,fund_result['rebalance_date'][-1]],[fund_result['rebalance_date'][-1],today]]
            _res_df = []
            for idx,dl in enumerate(date_list):
                b1 = dl[0]
                b2 = dl[1]
                _res = {}
                for fund_id, fund_pos_i in pos_df[b1].items():
                    if fund_pos_i['volume'] > 1:
                        if fund_pos_i['index_id'] not in _res:
                            _res[fund_pos_i['index_id']] = [fund_pos_i['fund_id']]
                        else:
                            _res[fund_pos_i['index_id']].append(fund_pos_i['fund_id'])
                if idx == 0:
                    _df = _index_price_ratio_item(_res, b1, b2, False, True)
                elif idx == 1:
                    _df = _index_price_ratio_item(_res, b1, b2, True, False)
                _res_df.append(_df)
            df_1 = _res_df[0]
            df_2 = _res_df[1]
            df_2 = df_1.iloc[-1] * df_2
            res_index_df = df_1.iloc[:-1].append(df_2, sort=True)
        AssetPainter.plot_index_price_ratio(res_index_df,'price_ratio')
        return res_df[['begin_date','end_date','desc_name','index_id','weight','rets']]

    @staticmethod
    def rolling_analysis(rolling_year, index_id, m, mv_df, fund_annual_ret, ret_list=[]):
        if '沪深300' in mv_df:
            mv_df = mv_df.drop('沪深300', axis=1)
        mv_df = mv_df.join(m.dts.index_price[[index_id]])
        mv_df = mv_df / mv_df.shift(1)
        annual_ret = (mv_df.rolling(window=242*rolling_year).agg(lambda x : x.prod() - 1))/rolling_year
        FundPainter.plot_rolling_analysis(annual_ret,fund_annual_ret,rolling_year,index_id,ret_list)
     
    @staticmethod
    def regular_rebalance_in_index_strategy(m, begin_date,end_date,index_id,rebalance_year,top_funds):
        dts = m.dts.trading_days.datetime
        fund_score_dic = m._score_manager.score_cache
        fund_info = m.dts.fund_info
        index_price = m.dts.index_price
        fund_nav = m.dts.fund_nav
        result_df, pos_df = HybridPainter.regular_rebalance_in_index_strategy(begin_date,end_date,index_id,rebalance_year,top_funds,dts,fund_score_dic,fund_info,index_price,fund_nav)
        return result_df, pos_df

    @staticmethod
    def mining_fac_importances(m,start_date,end_date,index_id,top_num,wind_class_2):
        '''
        index_id = 'csi500'
        start_date = datetime.date(2020,1,1)
        end_date = datetime.date(2020,8,10)
        top_num = 5
        wind_class_2 = ['被动指数型基金']
        ReportHelper.mining_fac_importances(m,start_date,end_date,index_id,top_num,wind_class_2)
        '''
        _dts = m.dts.trading_days.datetime
        _dts = _dts[(_dts >= start_date) & (_dts <= end_date)]
        _start_date = _dts.iloc[0]
        _end_date = _dts.iloc[-1]
        fund_info = m.dts.fund_info
        _fund_list = fund_info[(fund_info['index_id'] == index_id) & fund_info['wind_class_2'].isin(wind_class_2)].fund_id.tolist()
        _index_fund_pack = m.dts.index_fund_indicator_pack.loc[index_id].loc[_start_date:_end_date]
        _dts = _index_fund_pack.index.get_level_values(0).tolist()
        _col1 = m.dts.fund_nav.columns.tolist()
        fl = [i for i in _fund_list if i in _col1]
        _fund_nav = m.dts.fund_nav[fl].loc[_start_date:_end_date].dropna(axis=1)
        top_funds = (_fund_nav.iloc[-1] / _fund_nav.iloc[0]).sort_values(ascending=False).index.tolist()
        indicator_existed_fund = _index_fund_pack.loc[_start_date].index.tolist()
        top_funds = [i for i in top_funds if i in indicator_existed_fund][:top_num]
        cur_d = _index_fund_pack.loc[_start_date]
        cur_d_sta = cur_d.apply(lambda x: (x - x.mean() + 1e-6)/ (x.std() + 1e-6), axis=0)
        cur_d_sta['beta'] = cur_d['beta']
        cur_d_sta['year_length'] = cur_d['year_length']
        cur_d = cur_d_sta
        return pd.DataFrame(cur_d.rank(pct=True).loc[top_funds].mean(axis=0)).rename(columns={0:'rank'}).sort_values('rank',ascending=False)


import pandas as pd
import numpy as np
import os
import datetime
import traceback
from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util import Retry
from ...api.raw import RawDataApi
from ...view.raw_models import YahooIndexPrice
from .raw_data_helper import RawDataHelper


class WebRawDataDownloader(object):

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self._session = Session()
        self._session.mount('https://', HTTPAdapter(
            max_retries=Retry(total=10, backoff_factor=0.02),
        ))
        self._session.mount('http://', HTTPAdapter(
            max_retries=Retry(total=10, backoff_factor=0.02),
        ))

    def read_csv(self, csv_path):
        return pd.read_csv(csv_path, index_col=1)

    def fund_fee(self):
        pass

    def wind_fund_info(self, csv_path):
        # Update manually
        # Default file in company wechat is 'fundlist_wind_20202015.xlsx'
        df = pd.read_excel(csv_path)
        df.columns = ['wind_id', 'desc_name', 'full_name', 'start_date', 'end_date', 'benchmark', 'wind_class_1',
            'wind_class_2', 'currency', 'base_fund_id', 'is_structured', 'is_open', 'manager_id', 'company_id']
        df['start_date'] = df['start_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d'))
        df['end_date'] = df['end_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d')
            if isinstance(x, str) else datetime.datetime.strptime('2040-12-31', '%Y-%m-%d'))
        df['is_structured'] = df['is_structured'].map(lambda x: False if x == '否' else True)
        df['is_open'] = df['is_open'].map(lambda x: False if x == '否' else True)
        self._data_helper._upload_raw(df, 'wind_fund_info')

    def cm_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 汇率数据.xls
            # df = pd.read_excel(csv_path)
            # df = df[['日期','美元中间价','欧元中间价','日元中间价','美元CFETS','欧元CFETS','日元CFETS']]
            # df.columns = ['datetime','usd_central_parity_rate','eur_central_parity_rate','jpy_central_parity_rate',
            #               'usd_cfets','eur_cfets','jpy_cfets']

            # auto update
            # data from http://www.chinamoney.com.cn/chinese/bkccpr
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
            url = ('http://www.chinamoney.com.cn/dqs/rest/cm-u-bk-ccpr/CcprHisExcelNew'
                f'?startDate={start_date}&endDate={end_date}&currency=USD/CNY,EUR/CNY,100JPY/CNY')
            headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}
            filename = ''
            try:
                r = self._session.get(url, headers=headers, timeout=(15, 60))
                filename = os.path.join('/tmp', 'cm.xlsx')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except Exception as e:
                print(f'[cm_index_price] got error {e}')
                return False
            df = pd.read_excel(filename)
            # Drop last 2 lines in excel, which are
            #   数据来源：	中国货币网
            #   www.chinamoney.com.cn
            df.drop(df.index[-2:], inplace=True)
            df = df[['日期', 'USD/CNY', 'EUR/CNY', '100JPY/CNY']]
            df.columns = ['datetime', 'usd_central_parity_rate', 'eur_central_parity_rate', 'jpy_central_parity_rate']
            df['jpy_central_parity_rate'] = df['jpy_central_parity_rate'] / 100
            df = df.sort_values('datetime')
            self._data_helper._upload_raw(df, 'cm_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def cxindex_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 中证信用债指数历史数据.xls
            # df = pd.read_excel(csv_path)
            # df.columns = ['index_id', 'symbol', 'datetime', 'open', 'high', 'low', 'close', 'total_turnover', 'volume']
            # df['ret'] = df['close'] / df['close'].shift(1) - 1
            # df['index_id'] = 'credit_debt'
            # df = df.drop(['symbol'], axis=1)[:-2]

            # auto update
            # data from http://www.csindex.com.cn/zh-CN/indices/index-detail/H11073
            # File download url
            url = 'http://www.csindex.com.cn/uploads/file/autofile/perf/H11073perf.xls'
            filename = ''
            try:
                r = self._session.get(url, timeout=(15, 60))
                filename = os.path.join('/tmp', 'H11073perf.xls')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except Exception as e:
                print(f'[cxindex_index_price] got error {e}')
                return False
            df = pd.read_excel(filename)
            df = df[['日期Date', '收盘Close', '涨跌幅(%)Change(%)', '成交量（万元）Volume(10 thousand CNY)',
                '成交金额（元）Turnover']]
            df.columns = ['datetime', 'close', 'ret', 'volume', 'total_turnover']
            df['volume'] = df['volume'].values * 10000
            df['ret'] = df['ret'] / 100
            df = df[(df['datetime']>=start_date) & (df['datetime']<=end_date)].sort_values('datetime')
            df['open'] = float('Nan')
            df['high'] = float('Nan')
            df['low'] = float('Nan')
            df['index_id'] = 'credit_debt'
            self._data_helper._upload_raw(df, 'cxindex_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def yahoo_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 大类资产指数据.xlsx
            # df = pd.read_excel(csv_path)
            # df['datetime'] = df['date']
            # for c in ['sp500', 'dax30', 'n225']:
            #     df_i = df[['datetime']].copy()
            #     df_i['close'] = df[c].copy()
            #     df_i['ret'] = df_i['close'] / df_i['close'].shift(1) - 1
            #     df_i['open'] = float('Nan')
            #     df_i['high'] = float('Nan')
            #     df_i['low'] = float('Nan')
            #     df_i['volume'] = float('Nan')
            #     df_i['total_turnover'] = float('Nan')
            #     df_i['index_id'] = c
            #     self._data_helper._upload_raw(df_i, 'yahoo_index_price')

            # auto update
            # data from
            # sp500 https://finance.yahoo.com/quote/%5EGSPC/history?p=%5EGSPC
            # n225 https://finance.yahoo.com/quote/%5EN225/history?p=%5EN225
            # dax30 https://finance.yahoo.com/quote/%5EGDAXI/history/
            # File download url:
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGSPC?period1=1553667756&period2=1585290156&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EN225?period1=1553673330&period2=1585295730&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGDAXI?period1=1553673346&period2=1585295746&interval=1d&events=history

            url_prefix = 'https://query1.finance.yahoo.com/v7/finance/download/'

            trading_day = RawDataApi().get_em_tradedates().set_index('TRADEDATES')
            start_date_loc = trading_day.index.get_loc(pd.to_datetime(start_date).date(), method='ffill')
            end_date_loc = trading_day.index.get_loc(pd.to_datetime(end_date).date(), method='ffill')

            assert isinstance(start_date_loc, int), 'start date is not a valid trading day'
            assert isinstance(end_date_loc, int), 'end date is not a valid trading day'

            # TODO: (deprecated)由于我们目前更新的时间（次日0点），美股还没收盘，所以这里只能更新前一天的数据；之后再看看可能需要分批做自动更新
            # TODO: 更新时间暂时统一挪到美股收盘后了，这里直接更新当天数据
            extra_days = 0
            assert start_date_loc >= extra_days, 'there is no enough extra days before start date'
            start_date_with_extra_days = trading_day.index.array[start_date_loc - extra_days]
            print(f'start date with extra days: {start_date_with_extra_days}')

            assert end_date_loc >= extra_days, 'there is no enough extra days before end date'
            end_date_with_extra_days = trading_day.index.array[end_date_loc - extra_days]
            print(f'end date with extra days: {end_date_with_extra_days}')

            start_date = min(pd.to_datetime(start_date, infer_datetime_format=True) - datetime.timedelta(days=7), start_date_with_extra_days)
            start_timestamp = str(int((start_date).timestamp()))
            end_timestamp = str(int((datetime.datetime.strptime(end_date, '%Y%m%d') + datetime.timedelta(days=1)).timestamp()))
            url_postfix = f'?period1={start_timestamp}&period2={end_timestamp}&interval=1d&events=history'
            url_type_list = ['%5EGSPC', '%5EN225', '%5EGDAXI']
            csv_list = ['^GSPC.csv', '^N225.csv', '^GDAXI.csv']
            index_list = ['sp500', 'n225', 'dax30']
            df_list = []
            for url_type, csv_i, index_i in zip(url_type_list, csv_list, index_list):
                url = url_prefix + url_type + url_postfix
                filename = ''
                try:
                    r = self._session.get(url, timeout=(15, 60))
                    filename = os.path.join('/tmp', csv_i)
                    with open(filename, 'wb') as f:
                        f.write(r.content)
                except Exception as e:
                    print(f'[yahoo_index_price] got error {e}')
                    continue
                df = pd.read_csv(filename)
                df = df[['Date', 'Open', 'High', 'Low', 'Close', 'Adj Close', 'Volume']]
                df.columns = ['datetime', 'open', 'high', 'low', 'close', 'adjclose', 'volume']
                df['datetime'] = df['datetime'].map(lambda x: pd.to_datetime(x, infer_datetime_format=True).date())
                df['index_id'] = index_i
                df['total_turnover'] = float('Nan')
                df['ret'] = df['adjclose'] / df['adjclose'].shift(1) - 1
                df = df.drop(['adjclose'], axis=1)
                df = df[df['datetime'].between(start_date_with_extra_days, end_date_with_extra_days)].sort_values('datetime')
                if not df.empty:
                    df_list.append(df)
            if df_list:
                df_all = pd.concat(df_list)
                self._data_helper._upload_raw(df_all, YahooIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def download_all(self, start_date, end_date):
        failed_tasks = []
        if not self.cm_index_price(start_date, end_date):
            failed_tasks.append('cm_index_price')

        if not self.cxindex_index_price(start_date, end_date):
            failed_tasks.append('cxindex_index_price')

        if not self.yahoo_index_price(start_date, end_date):
            failed_tasks.append('yahoo_index_price')

        return failed_tasks


if __name__ == "__main__":
    data_helper = RawDataHelper()
    web_downloader = WebRawDataDownloader(data_helper)
    # web_downloader.cm_index_price('20200408', '20200508')
    # web_downloader.yahoo_index_price('20200427', '20200710')
